# -*- coding: utf-8 -*-
"""Database module of Stalker.

Whenever stalker.db or something under it imported, the :func:`stalker.db.setup` becomes
available to let one set up the database.
"""
import logging
import os
from typing import Any, Dict, List, Optional, Union

from sqlalchemy import Column, Table, Text, engine_from_config, text
from sqlalchemy.exc import IntegrityError, OperationalError, ProgrammingError

from stalker import (
    DateRangeMixin,
    Department,
    EntityType,
    Group,
    Permission,
    ReferenceMixin,
    Repository,
    ScheduleMixin,
    Status,
    StatusList,
    StatusMixin,
    Studio,
    Type,
    User,
    defaults,
    log,
)
from stalker.db.declarative import Base
from stalker.db.session import DBSession


logger: logging.Logger = log.get_logger(__name__)

# TODO: Try to get it from the API (it was not working inside a package before)
alembic_version: str = "9f9b88fef376"


def setup(settings: Optional[Dict[str, Any]] = None) -> None:
    """Connect the system to the given database.

    If the database is None then it sets up using the default database in the
    settings file.

    Args:
        settings (Dict[str, Any]): This is a dictionary which has keys prefixed
            with "sqlalchemy" and shows the settings. The most important one is
            the engine. The default is None, and in this case it uses the
            settings from stalker.config.Config.database_engine_settings
    """
    if settings is None:
        settings = defaults.database_engine_settings
        logger.debug("no settings given, using the default setting")

    # logger.debug(f"settings: {settings}")
    # create engine

    logger.debug(f"settings: {settings}")
    engine = engine_from_config(settings, "sqlalchemy.")

    logger.debug(f"engine: {engine}")

    # create the Session class
    DBSession.remove()
    DBSession.configure(bind=engine)

    # check alembic versions of the database
    # and raise an error if it is not matching with the system
    check_alembic_version()

    # create the database
    logger.debug("creating the tables")

    Base.metadata.create_all(engine)
    DBSession.commit()

    # update defaults
    update_defaults_with_studio()

    # create repo env variables
    create_repo_vars()


def update_defaults_with_studio() -> None:
    """Update the default values from Studio instance.

    Update only if a database and a Studio instance is present.
    """
    with DBSession.no_autoflush:
        studio = Studio.query.first()
        # studio = DBSession.query(Studio).first()
        logger.debug("studio: {}".format(studio))
        if studio:
            logger.debug("found a studio, updating defaults")
            studio.update_defaults()


def init() -> None:
    """Fill the database with default values."""
    logger.debug("initializing database")

    # register all Actions available for all SOM classes
    class_names = [
        "Asset",
        "AuthenticationLog",
        "Budget",
        "BudgetEntry",
        "Client",
        "Daily",
        "Department",
        "Entity",
        "EntityGroup",
        "File",
        "FilenameTemplate",
        "Good",
        "Group",
        "ImageFormat",
        "Invoice",
        "Message",
        "Note",
        "Page",
        "Payment",
        "Permission",
        "PriceList",
        "Project",
        "Repository",
        "Review",
        "Role",
        "Scene",
        "Sequence",
        "Shot",
        "SimpleEntity",
        "Status",
        "StatusList",
        "Structure",
        "Studio",
        "Tag",
        "Task",
        "Ticket",
        "TicketLog",
        "TimeLog",
        "Type",
        "User",
        "Vacation",
        "Variant",
        "Version",
    ]

    for class_name in class_names:
        _temp = __import__("stalker", globals(), locals(), [class_name], 0)
        class_ = eval("_temp.{}".format(class_name))
        register(class_)

    # create the admin if needed
    admin = None

    if defaults.auto_create_admin:
        admin = __create_admin__()

    # create statuses
    create_ticket_statuses()

    # create statuses for Tickets
    create_entity_statuses(
        entity_type="Daily",
        status_names=defaults.daily_status_names,
        status_codes=defaults.daily_status_codes,
        user=admin,
    )
    create_entity_statuses(
        entity_type="Project",
        status_names=defaults.project_status_names,
        status_codes=defaults.project_status_codes,
        user=admin,
    )
    create_entity_statuses(
        entity_type="Task",
        status_names=defaults.task_status_names,
        status_codes=defaults.task_status_codes,
        user=admin,
    )
    create_entity_statuses(
        entity_type="Review",
        status_names=defaults.review_status_names,
        status_codes=defaults.review_status_codes,
        user=admin,
    )

    # create alembic revision table
    create_alembic_table()

    logger.debug("finished initializing the database")


def create_repo_vars() -> None:
    """Create environment variables for all the repositories in the current database."""
    # get all the repositories
    all_repos = Repository.query.all()
    for repo in all_repos:
        os.environ[repo.env_var] = repo.path


def get_alembic_version() -> Union[None, str]:
    """Return the alembic version of the database.

    Returns:
        str: The alembic version.
    """
    # try to query the version value
    conn = DBSession.connection()
    engine = conn.engine
    if not engine.dialect.has_table(conn, "alembic_version"):
        return None

    sql_query = "select version_num from alembic_version"
    try:
        return DBSession.connection().execute(text(sql_query)).fetchone()[0]
    except (OperationalError, ProgrammingError, TypeError):
        DBSession.rollback()
        return None


def check_alembic_version() -> None:
    """Check the alembic version of the database.

    Raises:
        ValueError: If the alembic version is not matching with current version of
            Stalker.
    """
    current_alembic_version = get_alembic_version()
    logger.debug(f"current_alembic_version: {current_alembic_version}")
    if current_alembic_version and current_alembic_version != alembic_version:
        # invalidate the connection
        DBSession.connection().invalidate()

        # and raise a ValueError (which I'm not sure is the correct exception)
        raise ValueError(f"Please update the database to version: {alembic_version}")


def create_alembic_table() -> None:
    """Create the default alembic_version table.

    Also create the data so that any new database will be considered as the latest
    version.
    """
    # Now, this is not the correct way of doing this, there is a proper way of
    # doing it and it is explained nicely in the Alembic library documentation.
    #
    # But it is simply not working when Stalker is installed as a package.
    #
    # So as a workaround here we are doing it manually
    # don't forget to update the version_num (and the corresponding test
    # whenever a new alembic revision is created)

    version_num = alembic_version
    table_name = "alembic_version"
    conn = DBSession.connection()
    engine = conn.engine

    # check if the table is already there
    table = Table(
        table_name, Base.metadata, Column("version_num", Text), extend_existing=True
    )
    if not engine.dialect.has_table(conn, table_name):
        logger.debug("creating alembic_version table")

        # create the table no matter if it exists or not we need it either way
        Base.metadata.create_all(engine)

    # first try to query the version value
    sql_query = "select version_num from alembic_version"
    try:
        version_num = DBSession.connection().execute(text(sql_query)).fetchone()[0]
    except TypeError:
        logger.debug(f"inserting {version_num} to alembic_version table")
        # the table is there but there is no value so insert it
        ins = table.insert().values(version_num=version_num)
        DBSession.connection().execute(ins)
        DBSession.commit()
        logger.debug("alembic_version table is created and initialized")
    else:
        # the value is there do not touch the table
        logger.debug("alembic_version table is already there, not doing anything!")


def __create_admin__() -> User:
    """Create the admin.

    Returns:
        User: The admin user.
    """
    logger.debug("creating the default administrator user")

    # create the admin department
    admin_department = Department.query.filter_by(
        name=defaults.admin_department_name
    ).first()

    if not admin_department:
        admin_department = Department(name=defaults.admin_department_name)
        DBSession.add(admin_department)
        DBSession.commit()
        # create the admins group
    admins_group = Group.query.filter_by(name=defaults.admin_group_name).first()

    if not admins_group:
        admins_group = Group(name=defaults.admin_group_name)
        DBSession.add(admins_group)

    # check if there is already an admin in the database
    admin = User.query.filter_by(name=defaults.admin_name).first()
    if admin:
        # there should be an admin user do nothing
        logger.debug("there is an admin already")
        return admin
    else:
        admin = User(
            name=defaults.admin_name,
            login=defaults.admin_login,
            password=defaults.admin_password,
            email=defaults.admin_email,
            departments=[admin_department],
            groups=[admins_group],
        )

        DBSession.add(admin)
        DBSession.commit()

        # admin.created_by = admin
        # admin.updated_by = admin

        # update the department as created and updated by admin user
        admin_department.created_by = admin
        admin_department.updated_by = admin

        admins_group.created_by = admin
        admins_group.updated_by = admin

        DBSession.add(admin)
        DBSession.commit()

    return admin


def create_ticket_statuses() -> None:
    """Create the default ticket statuses."""
    # create as admin
    admin = User.query.filter(User.login == defaults.admin_name).first()

    # create statuses for Tickets
    ticket_names = defaults.ticket_status_names
    ticket_codes = defaults.ticket_status_codes
    create_entity_statuses("Ticket", ticket_names, ticket_codes, admin)

    # Again I hate doing this in this way
    types = Type.query.filter_by(target_entity_type="Ticket").all()
    t_names = [t.name for t in types]

    # create Ticket Types
    logger.debug("Creating Ticket Types")
    if "Defect" not in t_names:
        ticket_type_1 = Type(
            name="Defect",
            code="Defect",
            target_entity_type="Ticket",
            created_by=admin,
            updated_by=admin,
        )
        DBSession.add(ticket_type_1)

    if "Enhancement" not in t_names:
        ticket_type_2 = Type(
            name="Enhancement",
            code="Enhancement",
            target_entity_type="Ticket",
            created_by=admin,
            updated_by=admin,
        )
        DBSession.add(ticket_type_2)

    try:
        DBSession.commit()
    except IntegrityError:
        DBSession.rollback()
        logger.debug("Ticket Types are already in the database!")
    else:
        # DBSession.flush()
        logger.debug("Ticket Types are created successfully")


def create_entity_statuses(
    entity_type: str = "",
    status_names: Optional[List[str]] = None,
    status_codes: Optional[List[str]] = None,
    user: Optional[User] = None,
) -> None:
    """Create the default task statuses.

    Args:
        entity_type (str): The entity type.
        status_names (List[str]): The list of status names.
        status_codes (List[str]): The list of status codes in the same order of the
            ``status_names`` args.
        user (stalker.model.auth.User): The :class:`stalker.model.auth.User` instance
            to use as the creator for the newly created data.

    Raises:
        ValueError: If entity_type, status_names or status_codes are empty.
    """
    if not entity_type:
        DBSession.rollback()
        raise ValueError("Please supply entity_type")

    if not status_names:
        DBSession.rollback()
        raise ValueError("Please supply status names")

    if not status_codes:
        DBSession.rollback()
        raise ValueError("Please supply status codes")

    # create statuses for entity
    logger.debug(f"Creating {entity_type} Statuses")

    with DBSession.no_autoflush:
        statuses = Status.query.filter(Status.name.in_(status_names)).all()

    logger.debug(f"status_names: {status_names}")
    logger.debug(f"statuses: {statuses}")
    status_names_in_db = list(map(lambda x: x.name, statuses))
    logger.debug(f"statuses_names_in_db: {status_names_in_db}")

    for name, code in zip(status_names, status_codes):
        if name not in status_names_in_db:
            logger.debug(f"Creating Status: {name} ({code})")
            new_status = Status(name=name, code=code, created_by=user, updated_by=user)
            statuses.append(new_status)
            DBSession.add(new_status)
        else:
            logger.debug(f"Status {name} ({code}) is already created skipping!")

    # create the Status List
    status_list = StatusList.query.filter(
        StatusList.target_entity_type == entity_type
    ).first()

    if status_list is None:
        logger.debug(f"No {entity_type} Status List found, creating new!")
        status_list = StatusList(
            name=f"{entity_type} Statuses",
            target_entity_type=entity_type,
            created_by=user,
            updated_by=user,
        )
    else:
        logger.debug(f"{entity_type} Status List already created, updating statuses")

    status_list.statuses = statuses
    DBSession.add(status_list)

    try:
        DBSession.commit()
    except (IntegrityError, OperationalError) as e:
        logger.debug(f"error in DBSession.commit, rolling back: {e}")
        DBSession.rollback()
    else:
        logger.debug(f"Created {entity_type} Statuses successfully")
        DBSession.flush()


def register(class_: Type) -> None:
    """Register the given class to the database.

    It is mainly used to create the :class:`.Action` s needed for the
    :class:`.User` s and :class:`.Group` s to be able to interact with the
    given class. Whatever class you have created needs to be registered.

    Example, let's say that you have a data class which is specific to your
    studio and it is not present in Stalker Object Model (SOM), so you need to
    extend SOM with a new data type. Here is a simple Data class inherited from
    the :class:`.SimpleEntity` class (which is the simplest class you should
    inherit your classes from or use more complex classes down to the
    hierarchy)::

      from sqlalchemy import Column, Integer, ForeignKey
      from sqlalchemy.orm import Mapped, mapped_column
      from stalker.models.entity import SimpleEntity

      class MyDataClass(SimpleEntity):
        '''This is an example class holding a studio specific data which is not
        present in SOM.
        '''

        __tablename__ = 'MyData'
        __mapper_arguments__ = {'polymorphic_identity': 'MyData'}

        my_data_id : Mapped[int] = mapped_column(
            'id',
            ForeignKey('SimpleEntities.c.id'),
            primary_key=True,
        )

    Now because Stalker is using Pyramid authorization mechanism it needs to be
    able to have an :class:`.Permission` about your new class, so you can
    assign this :class;`.Permission` to your :class:`.User` s or
    :class:`.Group` s. So you need to register your new class with
    :func:`stalker.db.register` like shown below::

    .. code-block: python

        from stalker.db import setup
        setup.register(MyDataClass)

    This will create the necessary Actions in the 'Actions' table on your
    database, then you can create :class:`.Permission` s and assign these to
    your :class:`.User` s and :class:`.Group` s so they are Allowed or Denied
    to do the specified Action.

    Args:
        class_ (Type): The class itself that needs to be registered.

    Raises:
        TypeError: If the class_ arg is not a ``type`` instance.
    """
    # create the Permissions
    permissions_db = Permission.query.all()

    if not isinstance(class_, type):
        raise TypeError("To register a class please supply the class itself.")

    # register the class name to entity_types table
    class_name = class_.__name__
    if not EntityType.query.filter_by(name=class_name).first():
        new_entity_type = EntityType(class_name)
        # update attributes
        if issubclass(class_, StatusMixin):
            new_entity_type.statusable = True
        if issubclass(class_, DateRangeMixin):
            new_entity_type.dateable = True
        if issubclass(class_, ScheduleMixin):
            new_entity_type.schedulable = True
        if issubclass(class_, ReferenceMixin):
            new_entity_type.accepts_references = True

        DBSession.add(new_entity_type)

    for action in defaults.actions:
        for access in ["Allow", "Deny"]:
            permission_obj = Permission(access, action, class_name)
            if permission_obj not in permissions_db:
                DBSession.add(permission_obj)

    try:
        DBSession.commit()
    except IntegrityError:
        DBSession.rollback()



# -*- coding: utf-8 -*-
"""The Message related classes and functions are situated here."""
from typing import Any, Dict

from sqlalchemy import ForeignKey
from sqlalchemy.orm import Mapped, mapped_column

from stalker.log import get_logger
from stalker.models.entity import Entity
from stalker.models.mixins import StatusMixin

logger = get_logger(__name__)


class Message(Entity, StatusMixin):
    """The base of the messaging system in Stalker.

    Messages are one of the ways to collaborate in Stalker. The model of the
    messages is taken from the e-mail system. So it is pretty similar to an
    e-mail message.

    Args:
        from (User): The :class:`.User` object sending the message.
        to (User): The list of :class:`.User` s to receive this message.
        subject (str): The subject of the message.
        body (str): tThe body of the message.
        in_reply_to (Message): The :class:`.Message` object which this message is a
            reply to.
        replies (Message): The list of :class:`.Message` objects which are the direct
            replies of this message.
        attachments (SimpleEntity): A list of :class:`.SimpleEntity` objects attached to
            this message (so anything can be attached to a message).
    """

    __auto_name__ = True
    __tablename__ = "Messages"
    __mapper_args__ = {"polymorphic_identity": "Message"}
    message_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Entities.id"), primary_key=True
    )

    def __init__(self, **kwargs: Dict[str, Any]) -> None:
        super(Message, self).__init__(**kwargs)
        StatusMixin.__init__(self, **kwargs)



# -*- coding: utf-8 -*-
"""SimpleEntity, Entity, EntityGroup and other related functions are situated here."""

import functools
import re
import uuid
from datetime import datetime
from typing import Any, Dict, List, Optional, TYPE_CHECKING, Union

import pytz

from sqlalchemy import Column, ForeignKey, Integer, String, Table, Text
from sqlalchemy.orm import Mapped, mapped_column, relationship, validates

from stalker.db.declarative import Base
from stalker.db.types import GenericDateTime
from stalker.log import get_logger

logger = get_logger(__name__)

if TYPE_CHECKING:  # pragma: no cover
    from stalker.models.auth import User
    from stalker.models.file import File
    from stalker.models.note import Note
    from stalker.models.tag import Tag
    from stalker.models.type import Type


class SimpleEntity(Base):
    """The base class of all the others.

    The ``SimpleEntity`` is the starting point of the Stalker Object Model, it
    starts by adding the basic information about an entity which are
    :attr:`.name`, :attr:`.description`, the audit information like
    :attr:`.created_by`, :attr:`.updated_by`, :attr:`.date_created`,
    :attr:`.date_updated` and a couple of naming attributes like
    :attr:`.nice_name` and last but not least the :attr:`.type` attribute which
    is very important for entities that needs a type.

    .. versionadded: 0.2.2.3
        :attr:`.html_style` and :attr:`.html_class` attributes:

        SimpleEntity instances now have two new attributes called
        :attr:`.html_style` and :attr:`.html_class` which can be used to store
        html styles and html classes per entity. (Hint: Can be used to colorize
        different type of Tasks in different colors or different statused tasks
        in different classes etc.)

    .. note::

       For derived classes if the
       :attr:`.SimpleEntity.type` needed to be specifically specified, that is
       it cannot be None or nothing else then a :class:`.Type` instance, set
       the ``strictly_typed`` class attribute to True::

           class NewClass(SimpleEntity):
               __strictly_typed__ = True

       This will ensure that the derived class always have a proper
       :attr:`.SimpleEntity.type` attribute and cannot be initialized without
       one.

    Two SimpleEntities considered to be equal if they have the same
    :attr:`.name`, the other attributes doesn't matter.

    .. versionadded:: 0.2.0
       Name attribute can be skipped. Starting from version 0.2.0 the ``name``
       attribute can be skipped. For derived classes use the ``__auto_name__``
       class attribute to control auto naming behavior.

    Args:
        name (str): A string value that holds the name of this entity. It
            should not contain any white space at the beginning and at the end
            of the string. Valid characters are [a-zA-Z0-9_/S].

            Advanced::

                For classes derived from the SimpleEntity, if an automatic name
                is desired, the ``__auto_name__`` class attribute can be set to
                True. Then Stalker will automatically generate an uuid4
                sequence for the name attribute.

        description (str): A string attribute that holds the description of
            this entity object, it could be an empty string, and it could not
            again have white spaces at the beginning and at the end of the
            string, again any given objects will be converted to strings
        generic_text (str): A string attribute that holds any text based
            information that should be affiliated with this entity, it could be
            an empty string, and it could not again have white spaces at the
            beginning and at the end of the string, again any given objects
            will be converted to strings.
        created_by (User): The :class:`.User` who has created this object.
        updated_by (User): The :class:`.User` who has updated this object
            lastly. The created_by and updated_by attributes point the same
            object if this object is just created.
        date_created (datetime): The date that this object is created.
        date_updated (datetime): The date that this object is updated lastly.
            For newly created entities this is equal to date_created and the
            date_updated cannot point a date which is before date_created.
        type (Type): The type of the current SimpleEntity. Used across several
            places in Stalker. Can be None. The default value is None.
    """

    # auto generate name values
    __auto_name__ = True
    __strictly_typed__ = False

    # TODO: Allow the user to specify the formatting of the name attribute with
    #       a formatter function
    __name_formatter__ = None

    __tablename__ = "SimpleEntities"
    id: Mapped[int] = mapped_column(primary_key=True)

    entity_type: Mapped[str] = mapped_column(String(128), nullable=False)
    __mapper_args__ = {
        "polymorphic_on": entity_type,
        "polymorphic_identity": "SimpleEntity",
    }

    name: Mapped[str] = mapped_column(
        String(256), nullable=False, doc="Name of this object"
    )

    description: Mapped[Optional[str]] = mapped_column(
        Text, doc="Description of this object."
    )

    created_by_id: Mapped[Optional[int]] = mapped_column(
        ForeignKey("Users.id", use_alter=True, name="SimpleEntities_created_by_id_fkey"),
        doc="The id of the :class:`.User` who has created this entity.",
    )

    created_by: Mapped[Optional["User"]] = relationship(
        backref="entities_created",
        primaryjoin="SimpleEntity.created_by_id==User.user_id",
        doc="The :class:`.User` who has created this object.",
    )

    updated_by_id: Mapped[Optional[int]] = mapped_column(
        "updated_by_id",
        ForeignKey("Users.id", use_alter=True, name="SimpleEntities_updated_by_id_fkey"),
        nullable=True,
        doc="The id of the :class:`.User` who has updated this entity.",
    )

    updated_by: Mapped[Optional["User"]] = relationship(
        backref="entities_updated",
        primaryjoin="SimpleEntity.updated_by_id==User.user_id",
        post_update=True,
        doc="The :class:`.User` who has updated this object.",
    )

    date_created: Mapped[Optional[datetime]] = mapped_column(
        GenericDateTime,
        default=functools.partial(datetime.now, pytz.utc),
        doc="""A :class:`datetime` instance showing the creation date and time
        of this object.""",
    )

    date_updated: Mapped[Optional[datetime]] = mapped_column(
        GenericDateTime,
        default=functools.partial(datetime.now, pytz.utc),
        doc="""A :class:`datetime` instance showing the update date and time of
        this object.""",
    )

    type_id: Mapped[Optional[int]] = mapped_column(
        ForeignKey("Types.id", use_alter=True, name="SimpleEntities_type_id_fkey"),
        doc="""The id of the :class:`.Type` of this entity. Mainly used by
        SQLAlchemy to create a Many-to-One relates between SimpleEntities and
        Types.
        """,
    )

    type: Mapped[Optional["Type"]] = relationship(
        primaryjoin="SimpleEntities.c.type_id==Types.c.id",
        post_update=True,
        doc="""The type of the object.

        It is a :class:`.Type` instance with a proper
        :attr:`.Type.target_entity_type`.
        """,
    )

    generic_data: Mapped[Optional[List["SimpleEntity"]]] = relationship(
        secondary="SimpleEntity_GenericData",
        primaryjoin="SimpleEntities.c.id=="
        "SimpleEntity_GenericData.c.simple_entity_id",
        secondaryjoin="SimpleEntity_GenericData.c.other_simple_entity_id=="
        "SimpleEntities.c.id",
        post_update=True,
        doc="This attribute can hold any kind of data which exists in SOM.",
    )

    generic_text: Mapped[Optional[str]] = mapped_column(
        "generic_text", Text, doc="This attribute can hold any text."
    )

    thumbnail_id: Mapped[Optional[int]] = mapped_column(
        ForeignKey(
            "Files.id",
            use_alter=True,
            name="SimpleEntities_thumbnail_id_fkey",
        )
    )

    thumbnail: Mapped[Optional["File"]] = relationship(
        primaryjoin="SimpleEntities.c.thumbnail_id==Files.c.id",
        post_update=True,
    )

    html_style: Mapped[Optional[str]] = mapped_column(
        String(64), nullable=True, default=""
    )
    html_class: Mapped[Optional[str]] = mapped_column(
        String(64), nullable=True, default=""
    )

    stalker_version: Mapped[Optional[str]] = mapped_column(String(256))

    def __init__(
        self,
        name: Optional[str] = None,
        description: Optional[str] = "",
        generic_text: str = "",
        type: Optional["Type"] = None,
        created_by: Optional["User"] = None,
        updated_by: Optional["User"] = None,
        date_created: Optional[datetime] = None,
        date_updated: Optional[datetime] = None,
        thumbnail: Optional["File"] = None,
        html_style: Optional[str] = "",
        html_class: Optional[str] = "",
        **kwargs: Optional[Dict[str, Any]],
    ) -> None:  # noqa: W0613

        # name and nice_name
        self._nice_name = ""

        self.name = name

        self.description = description
        self.created_by = created_by
        self.updated_by = updated_by
        if date_created is None:
            date_created = datetime.now(pytz.utc)
        if date_updated is None:
            date_updated = date_created

        self.date_created = date_created
        self.date_updated = date_updated
        self.type = type
        self.thumbnail = thumbnail
        self.generic_text = generic_text
        self.html_style = html_style
        self.html_class = html_class

        import stalker

        self.stalker_version = stalker.__version__

    def __repr__(self) -> str:
        """Return the str representation of this SimpleEntity.

        Returns:
            str: The str representation of this SimpleEntity.
        """
        return f"<{self.name} ({self.entity_type})>"

    def __eq__(self, other: Any) -> bool:
        """Check equality.

        Args:
            other (Any): An object to check the equality of.

        Returns:
            bool: If the other is a SimpleEntity and its name equals to this one.
        """
        from stalker.db.session import DBSession

        with DBSession.no_autoflush:
            return isinstance(other, SimpleEntity) and self.name == other.name

    def __ne__(self, other: Any) -> bool:
        """Check inequality.

        This uses the __eq__ operator to get the inequality.

        Args:
            other (Any): An object to check the inequality of.

        Returns:
            bool: True if other is not equal to this instance.
        """
        return not self.__eq__(other)

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return hash("{}:{}:{}".format(self.id, self.name, self.entity_type))

    @validates("description")
    def _validate_description(self, key: str, description: str) -> str:
        """Validate the given description value.

        Args:
            key (str): The name of the validated column.
            description (str): The description value to be validated.

        Raises:
            TypeError: If the description is not None and not a str.

        Returns:
            str: The validated description value.
        """
        if description is None:
            description = ""

        if not isinstance(description, str):
            raise TypeError(
                "{}.description should be a string, not {}: '{}'".format(
                    self.__class__.__name__, description.__class__.__name__, description
                )
            )
        return description

    @validates("generic_text")
    def _validate_generic_text(self, key: str, generic_text: str) -> str:
        """Validate the given generic_text value.

        Args:
            key (str): The name of the validated column.
            generic_text (str): The generic_text value to be validated.

        Raises:
            TypeError: If the generic_text is not None and not a str.

        Returns:
            str: The validated generic_text value.
        """
        if generic_text is None:
            generic_text = ""

        if not isinstance(generic_text, str):
            raise TypeError(
                "{}.generic_text should be a string, not {}: '{}'".format(
                    self.__class__.__name__,
                    generic_text.__class__.__name__,
                    generic_text,
                )
            )
        return generic_text

    @validates("name")
    def _validate_name(self, key: str, name: str) -> str:
        """Validate the name value.

        Args:
            key (str): The name of the validated column.
            name (str): The name value to be validated.

        Raises:
            TypeError: If the name is not a str.
            ValueError: If the name becomes an empty str after formatting.

        Returns:
            str: The validated name value.
        """
        if self.__auto_name__:
            if name is None or name == "":
                # generate a uuid4
                name = "{}_{}".format(
                    self.__class__.__name__,
                    uuid.uuid4().urn.split(":")[2],
                )

        # it is None
        if name is None:
            raise TypeError(f"{self.__class__.__name__}.name cannot be None")

        if not isinstance(name, str):
            raise TypeError(
                f"{self.__class__.__name__}.name should be a string, "
                f"not {name.__class__.__name__}: '{name}'"
            )

        name = self._format_name(name)

        # it is empty
        if name == "":
            raise ValueError(
                f"{self.__class__.__name__}.name cannot be an empty string"
            )

        # also set the nice_name
        self._nice_name = self._format_nice_name(name)

        return name

    @classmethod
    def _format_name(cls, name: str) -> str:
        """Format the name value.

        Args:
            name (str): The name value.

        Returns:
            str: The formatted name value.
        """
        # remove unnecessary characters from the string
        name = name.strip()

        # remove multiple spaces
        name = re.sub(r"\s+", " ", name)

        return name

    @classmethod
    def _format_nice_name(cls, nice_name: str) -> str:
        """Format the given nice name value.

        Args:
            nice_name (str): The nice_name value to be formatted.

        Returns:
            str: The formatted nice name.
        """
        # remove unnecessary characters from the string
        nice_name = nice_name.strip()
        nice_name = re.sub(r"([^a-zA-Z0-9\s_\-@]+)", "", nice_name).strip()

        # remove all the characters which are not alphabetic from the start of
        # the string
        nice_name = re.sub(r"(^[^a-zA-Z0-9]+)", "", nice_name)

        # remove multiple spaces
        nice_name = re.sub(r"\s+", " ", nice_name)

        # # replace camel case letters
        # nice_name = re.sub(r"(.+?[a-z]+)([A-Z])", r"\1_\2", nice_name)

        # replace white spaces and dashes with underscore
        nice_name = re.sub("([ -])+", r"_", nice_name)

        # remove multiple underscores
        nice_name = re.sub(r"(_+)", r"_", nice_name)

        return nice_name

    @property
    def nice_name(self) -> str:
        """Nice name of this object.

        It has the same value with the name (contextually) but with a different
        format like, all the white spaces replaced by underscores ("_"), all the
        CamelCase form will be expanded by underscore (_) characters, and it is always
        lower case.

        Returns:
            str: The nice name value.
        """
        # also set the nice_name
        # if self._nice_name is None or self._nice_name == "":
        self._nice_name = self._format_nice_name(self.name)
        return self._nice_name

    @validates("created_by")
    def _validate_created_by(
        self, key: str, created_by: Union[None, "User"]
    ) -> Union[None, "User"]:
        """Validate the given created_by value.

        Args:
            key (str): The name of the validated column.
            created_by (Union[None, User]): The created_by value to be validated.

        Raises:
            TypeError: If the created_by value is not None and not a
                :class:`stalker.models.auth.User` instance.

        Returns:
            Union[None, User]: The validated created_by value.
        """
        from stalker.models.auth import User

        if created_by is not None:
            if not isinstance(created_by, User):
                raise TypeError(
                    f"{self.__class__.__name__}.created_by should be a "
                    "stalker.models.auth.User instance, "
                    f"not {created_by.__class__.__name__}: '{created_by}'"
                )
        return created_by

    @validates("updated_by")
    def _validate_updated_by(
        self, key: str, updated_by: Union[None, "User"]
    ) -> Union[None, "User"]:
        """Validate the given updated_by value.

        Args:
            key (str): The name of the validated column.
            updated_by (Union[None, User]): The updated_by value to be validated.

        Raises:
            TypeError: If the updated_by value is not None and not a
                :class:`stalker.models.auth.User` instance.

        Returns:
            Union[None, User]: The validated updated_by value.
        """
        from stalker.models.auth import User

        if updated_by is None:
            # set it to what created_by attribute has
            updated_by = self.created_by

        if updated_by is not None:
            if not isinstance(updated_by, User):
                raise TypeError(
                    f"{self.__class__.__name__}.updated_by should be a "
                    "stalker.models.auth.User instance, "
                    f"not {updated_by.__class__.__name__}: '{updated_by}'"
                )
        return updated_by

    @validates("date_created")
    def _validate_date_created(self, key: str, date_created: datetime) -> datetime:
        """Validate the given date_created value.

        Args:
            key (str): The name of the validated column.
            date_created (datetime): The value to be validated.

        Raises:
            TypeError: If the given date_created value is None or not a datetime
                instance.

        Returns:
            datetime: The validated date_created value.
        """
        if date_created is None:
            raise TypeError(f"{self.__class__.__name__}.date_created cannot be None")

        if not isinstance(date_created, datetime):
            raise TypeError(
                f"{self.__class__.__name__}.date_created should be a "
                "datetime.datetime instance, "
                f"not {date_created.__class__.__name__}: '{date_created}'"
            )

        return date_created

    @validates("date_updated")
    def _validate_date_updated(self, key: str, date_updated: datetime) -> datetime:
        """Validate the given date_updated.

        Args:
            key (str): The name of the validated column.
            date_updated (datetime): The date_updated to be validated.

        Raises:
            TypeError: If the date_updated value is ``None`` or date_updated is
                not a ``datetime`` instance.
            ValueError: If the date_updated is before than the date_created.

        Returns:
            datetime: The validated datetime_updated value.
        """
        # it is None
        if date_updated is None:
            raise TypeError(f"{self.__class__.__name__}.date_updated cannot be None")

        # it is not a datetime instance
        if not isinstance(date_updated, datetime):
            raise TypeError(
                f"{self.__class__.__name__}.date_updated should be a "
                "datetime.datetime instance, "
                f"not {date_updated.__class__.__name__}: '{date_updated}'"
            )

        # lower than date_created
        if date_updated < self.date_created:
            raise ValueError(
                "{class_name}.date_updated could not be set to a date before "
                "{class_name}.date_created, try setting the ``date_created`` "
                "first.".format(class_name=self.__class__.__name__)
            )
        return date_updated

    @validates("type")
    def _validate_type(self, key: str, type_: "Type") -> "Type":
        """Validate the given type value.

        Args:
            key (str): The name of the validated column.
            type_ (Type): The type value to be validated.

        Raises:
            TypeError: If this class is a strictly typed class and the type_ is not
                None and not a Type instance.

        Returns:
            Type: The validated type_ value.
        """
        if self.__strictly_typed__ or type_ is not None:
            from stalker.models.type import Type

            if not isinstance(type_, Type):
                raise TypeError(
                    f"{self.__class__.__name__}.type must be a "
                    "stalker.models.type.Type instance, "
                    f"not {type_.__class__.__name__}: '{type_}'"
                )
        return type_

    @validates("thumbnail")
    def _validate_thumbnail(self, key: str, thumb: "File") -> "File":
        """Validate the given thumb value.

        Args:
            key (str): The name of the validated column.
            thumb (File): The thumb value to be validated.

        Raises:
            TypeError: If the given thumb value is not None and not a File
                instance.

        Returns:
            Union[None, File]: The validated thumb value.
        """
        if thumb is not None:
            from stalker import File

            if not isinstance(thumb, File):
                raise TypeError(
                    f"{self.__class__.__name__}.thumbnail should be a "
                    "stalker.models.file.File instance, "
                    f"not {thumb.__class__.__name__}: '{thumb}'"
                )
        return thumb

    @property
    def tjp_id(self) -> str:
        """Return TaskJuggler compatible id.

        Returns:
            str: The TaskJuggler compatible id.
        """
        return f"{self.__class__.__name__}_{self.id}"

    @property
    def to_tjp(self) -> str:
        """Render a TaskJuggler compliant str used for TaskJuggler integration.

        Needs to be overridden in inherited classes.

        Raises:
            NotImplementedError: Always.
        """
        raise NotImplementedError(
            f"This property is not implemented in {self.__class__.__name__}"
        )

    @validates("html_style")
    def _validate_html_style(self, key: str, html_style: str) -> str:
        """Validate the given html_style value.

        Args:
            key (str): The name of the validated column.
            html_style (str): The html_style to be validated.

        Raises:
            TypeError: If the given html_style is not a str.

        Returns:
            str: The validated html_style value.
        """
        if html_style is None:
            html_style = ""

        if not isinstance(html_style, str):
            raise TypeError(
                f"{self.__class__.__name__}.html_style should be a str, "
                f"not {html_style.__class__.__name__}: '{html_style}'"
            )
        return html_style

    @validates("html_class")
    def _validate_html_class(self, key: str, html_class: str) -> str:
        """Validate the given html_class value.

        Args:
            key (str): The name of the validated column.
            html_class (str): The html_class to be validated.

        Raises:
            TypeError: If the html_class is not a str.

        Returns:
            str: The validated html_class value.
        """
        if html_class is None:
            html_class = ""

        if not isinstance(html_class, str):
            raise TypeError(
                f"{self.__class__.__name__}.html_class should be a str, "
                f"not {html_class.__class__.__name__}: '{html_class}'"
            )
        return html_class


class Entity(SimpleEntity):
    """Another base data class that adds tags and notes to the attributes list.

    This is the entity class which is derived from the SimpleEntity and adds
    only tags to the list of parameters.

    Two Entities considered equal if they have the same name. It doesn't matter
    if they have different tags or notes.

    Args:
        tags (List[Tag]): A list of :class:`.Tag` objects related to this entity.
            Tags could be an empty list, or when omitted it will be set to an
            empty list.
        notes (List[Note]): A list of :class:`.Note` instances. Can be an empty
            list, or when omitted it will be set to an empty list, when set to
            None it will be converted to an empty list.
    """

    __auto_name__ = True
    __tablename__ = "Entities"
    __mapper_args__ = {"polymorphic_identity": "Entity"}
    entity_id: Mapped[int] = mapped_column(
        "id", ForeignKey("SimpleEntities.id"), primary_key=True
    )

    tags: Mapped[Optional[List["Tag"]]] = relationship(
        "Tag",
        secondary="Entity_Tags",
        backref="entities",
        doc="""A list of tags attached to this object.

        It is a list of :class:`.Tag` instances which shows
        the tags of this object""",
    )

    notes: Mapped[Optional[List["Note"]]] = relationship(
        "Note",
        secondary="Entity_Notes",
        backref="entities",
        doc="""All the :class:`.Notes` s attached to this entity.

        It is a list of :class:`.Note` instances or an
        empty list, setting it to None will raise a TypeError.
        """,
    )

    def __init__(
        self, tags: Optional[List["Tag"]] = None, notes=None, **kwargs
    ) -> None:
        super(Entity, self).__init__(**kwargs)

        if tags is None:
            tags = []

        if notes is None:
            notes = []

        self.tags = tags
        self.notes = notes

    @validates("notes")
    def _validate_notes(self, key: str, note: "Note") -> "Note":
        """Validate the given note value.

        Args:
            key (str): The name of the validated column.
            note (Note): The note value to be validated.

        Raises:
            TypeError: If the given note value is not a Note instance.

        Returns:
            Note: The validated note value.
        """
        from stalker.models.note import Note

        if not isinstance(note, Note):
            raise TypeError(
                f"{self.__class__.__name__}.note should be a stalker.models.note.Note "
                f"instance, not {note.__class__.__name__}: '{note}'"
            )
        return note

    @validates("tags")
    def _validate_tags(self, key: str, tag: "Tag") -> "Tag":
        """Validate the given tag value.

        Args:
            key (str): The name of the validated column.
            tag (Tag): The tag value to be validated.

        Raises:
            TypeError: If the given tag value is not a Tag instance.

        Returns:
            Tag: The validated tag value.
        """
        from stalker.models.tag import Tag

        if not isinstance(tag, Tag):
            raise TypeError(
                f"{self.__class__.__name__}.tag should be a stalker.models.tag.Tag "
                f"instance, not {tag.__class__.__name__}: '{tag}'"
            )
        return tag

    def __eq__(self, other: Any) -> bool:
        """Check if the other object is equal to this one.

        Args:
            other (Any): An object.

        Returns:
            bool: True if the other object is also an Entity instance and has the same
                basic attribute values.
        """
        return super(Entity, self).__eq__(other) and isinstance(other, Entity)

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(Entity, self).__hash__()


class EntityGroup(Entity):
    """Groups a wide variety of objects together to let one easily reach them.

    :class:`.EntityGroup` helps to group different types of entities together to let one
    easily reach to them.
    """

    __auto_name__ = True
    __tablename__ = "EntityGroups"
    __mapper_args__ = {"polymorphic_identity": "EntityGroup"}
    entity_group_id: Mapped[int] = mapped_column(
        "id",
        ForeignKey("Entities.id"),
        primary_key=True,
    )

    entities: Mapped[Optional[List[SimpleEntity]]] = relationship(
        secondary="EntityGroup_Entities",
        post_update=True,
        backref="entity_groups",
        doc="All the :class:`.SimpleEntity`s grouped in this EntityGroup.",
    )

    def __init__(
        self,
        entities: Optional[List[Entity]] = None,
        **kwargs: Optional[Dict[str, Any]],
    ) -> None:
        super(Entity, self).__init__(**kwargs)

        if entities is None:
            entities = []

        self.entities = entities

    @validates("entities")
    def _validate_entities(self, key: str, entity: SimpleEntity) -> SimpleEntity:
        """Validate the given entity value.

        Args:
            key (str): The name of the validated column.
            entity (SimpleEntity): The entity value to be validated.

        Raises:
            TypeError: If the entity is not a SimpleEntity instance.

        Returns:
            SimpleEntity: The validated entity value.
        """
        if not isinstance(entity, SimpleEntity):
            raise TypeError(
                f"{self.__class__.__name__}.entities should be a list of "
                f"SimpleEntities, not {entity.__class__.__name__}: '{entity}'"
            )

        return entity

    def __eq__(self, other: Any) -> bool:
        """Check if the other is equal to this instance.

        Args:
            other (Any): The other EntityGroup to check the equality of.

        Returns:
            bool: True if the other is also a EntityGroup instance and has the same
                attribute values.
        """
        return (
            super(EntityGroup, self).__eq__(other)
            and isinstance(other, EntityGroup)
            and self.entities == other.entities
        )

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(EntityGroup, self).__hash__()


# Entity Tags
Entity_Tags = Table(
    "Entity_Tags",
    Base.metadata,
    Column(
        "entity_id",
        Integer,
        ForeignKey("Entities.id"),
        primary_key=True,
    ),
    Column(
        "tag_id",
        Integer,
        ForeignKey("Tags.id"),
        primary_key=True,
    ),
)

# Entity Notes
Entity_Notes = Table(
    "Entity_Notes",
    Base.metadata,
    Column(
        "entity_id",
        Integer,
        ForeignKey("Entities.id"),
        primary_key=True,
    ),
    Column(
        "note_id",
        Integer,
        ForeignKey("Notes.id"),
        primary_key=True,
    ),
)

# SimpleEntity Generic Data
SimpleEntity_GenericData = Table(
    "SimpleEntity_GenericData",
    Base.metadata,
    Column(
        "simple_entity_id", Integer, ForeignKey("SimpleEntities.id"), primary_key=True
    ),
    Column(
        "other_simple_entity_id",
        Integer,
        ForeignKey("SimpleEntities.id"),
        primary_key=True,
    ),
)

# EntityGroup Entities
EntityGroup_Entities = Table(
    "EntityGroup_Entities",
    Base.metadata,
    Column("entity_group_id", Integer, ForeignKey("EntityGroups.id"), primary_key=True),
    Column(
        "other_entity_id", Integer, ForeignKey("SimpleEntities.id"), primary_key=True
    ),
)



# -*- coding: utf-8 -*-
"""Config related functions and classes are situated here."""
import datetime
import os
import sys
from typing import Any, Dict

from stalker import log

logger = log.get_logger(__name__)


class ConfigBase(object):
    """Config abstraction.

    This is based on Sphinx's config idiom.
    """

    default_config_values: Dict[str, Any] = {}

    def __init__(self) -> None:
        self.config_values = self.default_config_values.copy()
        self.user_config: Dict[str, Any] = {}
        self._parse_settings()

    def _parse_settings(self) -> None:
        """Parse the settings.

        The priority order is:

            stalker.config
            config.py under .stalker_rc directory
            config.py under $STALKER_PATH

        Raises:
            RuntimeError: If there is a Syntax error in the configuration.
        """
        # for now just use $STALKER_PATH
        # try to get the environment variable
        if self.env_key not in os.environ:
            # don't do anything
            logger.debug("no environment key found for user settings")
        else:
            logger.debug("environment key found")

            resolved_path = os.path.expanduser(
                os.path.join(os.environ[self.env_key], "config.py")
            )

            # using `while` is not safe to expand variables
            # so expand vars for 100 times which already is ridiculously
            # complex
            max_recursion = 100
            i = 0
            while "$" in resolved_path and i < max_recursion:
                resolved_path = os.path.expandvars(resolved_path)
                i += 1

            try:
                logger.debug("importing user config")
                with open(resolved_path) as f:
                    exec(f.read(), self.user_config)
            except IOError:
                logger.warning(
                    f"The $STALKER_PATH: {resolved_path} doesn't exists! "
                    "skipping user config"
                )
            except SyntaxError as e:
                raise RuntimeError(
                    f"There is a syntax error in your configuration file: {e}"
                )
            finally:
                # append the data to the current settings
                logger.debug("updating system config")
                for key in self.user_config:
                    # if key in self.config_values:
                    self.config_values[key] = self.user_config[key]

    def __getattr__(self, name: str) -> Any:
        """Return the config value as if it is an attribute look up.

        Args:
            name (str): The name of the config value.

        Returns:
            Any: The value related to the given config value.
        """
        return self.config_values[name]

    def __getitem__(self, name: str) -> Any:
        """Return item with the key.

        Args:
            name (str): The key to find the value of.

        Returns:
            Any: The value related to the given key.
        """
        return getattr(self, name)

    def __setitem__(self, name: str, value: Any) -> None:
        """Set the item with index of name to value.

        Args:
            name (str): The name as the index.
            value (Any): The value to set the item to.
        """
        self.config_values[name] = value

    def __delitem__(self, name: str) -> None:
        """Delete the item with the given name.

        Args:
            name (str): The name of the item to delete.
        """
        self.config_values.pop(name)

    def __contains__(self, name: str) -> bool:
        """Check if this contains the name.

        Args:
            name (str): The config name.

        Returns:
            bool: True if this contains the name, False otherwise.
        """
        return name in self.config_values


class Config(ConfigBase):
    """Holds system-wide configuration variables.

    See `configuring stalker`_ for more detail.

    .. _configuring stalker: ../configure.html
    """

    env_key = "STALKER_PATH"

    default_config_values = dict(
        #
        # The default settings for the database, see sqlalchemy.create_engine
        # for possible parameters
        #
        database_engine_settings={
            "sqlalchemy.url": "sqlite://",
            "sqlalchemy.echo": False,
            # "sqlalchemy.pool_pre_ping": True,
        },
        database_session_settings={},
        # Local storage path
        local_storage_path=os.path.expanduser("~/.strc"),
        local_session_data_file_name="local_session_data",
        # Storage for uploaded files
        server_side_storage_path=os.path.expanduser("~/Stalker_Storage"),
        repo_env_var_template="REPO{code}",
        #
        # Tells Stalker to create an admin by default
        #
        auto_create_admin=True,
        #
        # these are for new projects
        # after creating the project you can change them from the interface
        #
        admin_name="admin",
        admin_login="admin",
        admin_password="admin",
        admin_email="admin@admin.com",
        admin_department_name="admins",
        admin_group_name="admins",
        # the default keyword which is going to be used in password scrambling
        key="stalker_default_key",
        actions=["Create", "Read", "Update", "Delete", "List"],  # CRUDL
        # Tickets
        ticket_label="Ticket",
        # define the available actions per Status
        ticket_status_names=["New", "Accepted", "Assigned", "Reopened", "Closed"],
        ticket_status_codes=["NEW", "ACP", "ASG", "ROP", "CLS"],
        ticket_resolutions=[
            "fixed",
            "invalid",
            "wontfix",
            "duplicate",
            "worksforme",
            "cantfix",
        ],
        ticket_workflow={
            "resolve": {
                "New": {"new_status": "Closed", "action": "set_resolution"},
                "Accepted": {"new_status": "Closed", "action": "set_resolution"},
                "Assigned": {"new_status": "Closed", "action": "set_resolution"},
                "Reopened": {"new_status": "Closed", "action": "set_resolution"},
            },
            "accept": {
                "New": {"new_status": "Accepted", "action": "set_owner"},
                "Accepted": {"new_status": "Accepted", "action": "set_owner"},
                "Assigned": {"new_status": "Accepted", "action": "set_owner"},
                "Reopened": {"new_status": "Accepted", "action": "set_owner"},
            },
            "reassign": {
                "New": {"new_status": "Assigned", "action": "set_owner"},
                "Accepted": {"new_status": "Assigned", "action": "set_owner"},
                "Assigned": {"new_status": "Assigned", "action": "set_owner"},
                "Reopened": {"new_status": "Assigned", "action": "set_owner"},
            },
            "reopen": {
                "Closed": {"new_status": "Reopened", "action": "del_resolution"}
            },
        },
        # Task Management
        timing_resolution=datetime.timedelta(hours=1),
        task_priority=500,
        working_hours={
            "mon": [[540, 1080]],  # 9:00 - 18:00
            "tue": [[540, 1080]],  # 9:00 - 18:00
            "wed": [[540, 1080]],  # 9:00 - 18:00
            "thu": [[540, 1080]],  # 9:00 - 18:00
            "fri": [[540, 1080]],  # 9:00 - 18:00
            "sat": [],  # saturday off
            "sun": [],  # sunday off
        },
        # this is strongly related with the working_hours settings,
        # this should match each other
        daily_working_hours=9,
        weekly_working_days=5,
        weekly_working_hours=45,
        yearly_working_days=261,  # math.ceil(5 * 52.1428)
        day_order=["mon", "tue", "wed", "thu", "fri", "sat", "sun"],
        datetime_units=["min", "h", "d", "w", "m", "y"],
        datetime_unit_names=["minute", "hour", "day", "week", "month", "year"],
        datetime_units_to_timedelta_kwargs={
            "min": {"name": "minutes", "multiplier": 1},
            "h": {"name": "hours", "multiplier": 1},
            "d": {"name": "days", "multiplier": 1},
            "w": {"name": "weeks", "multiplier": 1},
            "m": {"name": "days", "multiplier": 30},
            "y": {"name": "days", "multiplier": 365},
        },
        task_status_names=[
            "Waiting For Dependency",
            "Ready To Start",
            "Work In Progress",
            "Pending Review",
            "Has Revision",
            "Dependency Has Revision",
            "On Hold",
            "Stopped",
            "Completed",
        ],
        task_status_codes=[
            "WFD",
            "RTS",
            "WIP",
            "PREV",
            "HREV",
            "DREV",
            "OH",
            "STOP",
            "CMPL",
        ],
        project_status_names=["Ready To Start", "Work In Progress", "Completed"],
        project_status_codes=["RTS", "WIP", "CMPL"],
        review_status_names=["New", "Requested Revision", "Approved"],
        review_status_codes=["NEW", "RREV", "APP"],
        daily_status_names=["Open", "Closed"],
        daily_status_codes=["OPEN", "CLS"],
        task_schedule_models=["effort", "length", "duration"],
        task_dependency_gap_models=["length", "duration"],
        task_dependency_targets=["onend", "onstart"],
        allocation_strategy=[
            "minallocated",
            "maxloaded",
            "minloaded",
            "order",
            "random",
        ],
        persistent_allocation=True,
        tjp_main_template2="""# Generated By Stalker v{{stalker.__version__}}
{{studio.to_tjp}}

# resources
resource resources "Resources" {
{%- for vacation in studio.vacations %}
{{vacation.to_tjp}}
{%- endfor %}
{%- for user in studio.users %}
{{user.to_tjp}}
{%- endfor %}
}

# tasks
{{ tasks_buffer }}

# reports
taskreport breakdown "{{csv_file_name}}"{
    formats csv
    timeformat "%Y-%m-%d-%H:%M"
    columns id, start, end {%- if compute_resources %}, resources{% endif %}
}""",
        tj_command="tj3" if sys.platform == "win32" else "/usr/local/bin/tj3",
        path_template="{{project.code}}/{%- for parent_task in parent_tasks -%}{{parent_task.nice_name}}/{%- endfor -%}",  # noqa: B950
        filename_template='{{version.nice_name}}_r{{"%02d"|format(version.revision_number)}}_v{{"%03d"|format(version.version_number)}}',  # noqa: B950
        # --------------------------------------------
        # the following settings came from oyProjectManager
        sequence_format="%h%p%t %R",
        file_size_format="%.2f MB",
        date_time_format="%Y.%m.%d %H:%M",
        resolution_presets={
            "PC Video": [640, 480, 1.0],
            "NTSC": [720, 486, 0.91],
            "NTSC 16:9": [720, 486, 1.21],
            "PAL": [720, 576, 1.067],
            "PAL 16:9": [720, 576, 1.46],
            "HD 720": [1280, 720, 1.0],
            "HD 1080": [1920, 1080, 1.0],
            "1K Super 35": [1024, 778, 1.0],
            "2K Super 35": [2048, 1556, 1.0],
            "4K Super 35": [4096, 3112, 1.0],
            "A4 Portrait": [2480, 3508, 1.0],
            "A4 Landscape": [3508, 2480, 1.0],
            "A3 Portrait": [3508, 4960, 1.0],
            "A3 Landscape": [4960, 3508, 1.0],
            "A2 Portrait": [4960, 7016, 1.0],
            "A2 Landscape": [7016, 4960, 1.0],
            "50x70cm Poster Portrait": [5905, 8268, 1.0],
            "50x70cm Poster Landscape": [8268, 5905, 1.0],
            "70x100cm Poster Portrait": [8268, 11810, 1.0],
            "70x100cm Poster Landscape": [11810, 8268, 1.0],
            "1k Square": [1024, 1024, 1.0],
            "2k Square": [2048, 2048, 1.0],
            "3k Square": [3072, 3072, 1.0],
            "4k Square": [4096, 4096, 1.0],
        },
        default_resolution_preset="HD 1080",
        project_structure="""{% for shot in project.shots %}
                Shots/{{shot.code}}
                Shots/{{shot.code}}/Plate
                Shots/{{shot.code}}/Reference
                Shots/{{shot.code}}/Texture
            {% endfor %}
        {% for asset in project.assets%}
            {% set asset_path = project.full_path + '/Assets/' + asset.type.name + '/' + asset.code %}
            {{asset_path}}/Texture
            {{asset_path}}/Reference
        {% endfor %}
        """,  # noqa: B950
        thumbnail_format="jpg",
        thumbnail_quality=70,
        thumbnail_size=[320, 180],
    )



# -*- coding: utf-8 -*-
"""FilenameTemplate related functions and classes are situated here."""
from typing import Any, Dict, Optional, Union

from sqlalchemy import ForeignKey, Text
from sqlalchemy.orm import Mapped, mapped_column, validates

from stalker.log import get_logger
from stalker.models.entity import Entity
from stalker.models.mixins import TargetEntityTypeMixin

logger = get_logger(__name__)


class FilenameTemplate(Entity, TargetEntityTypeMixin):
    """Holds templates for filename and path conventions.

    FilenameTemplate objects help to specify where to place a :class:`.Version`
    related file.

    Although, it is mainly used by Stalker to define :class:`.Version` related
    file paths and file names to place them in to proper places inside a
    :class:`.Project`'s :attr:`.Project.structure`, the idea behind is open to
    endless possibilities.

    Here is an example::

        p1 = Project(name="Test Project") # shortened for this example

        # shortened for this example
        s1 = Structure(name="Commercial Project Structure")

        # this is going to be used by Stalker to decide the :stalker:`.File`
        # :stalker:`.File.filename` and :stalker:`.File.path` (which is the way
        # Stalker links external files to Version instances)
        f1 = FilenameTemplate(
            name="Asset Version Template",
            target_entity_type="Asset",
            path='$REPO{{project.repository.id}}/{{project.code}}/{%- for parent_task in parent_tasks -%}{{parent_task.nice_name}}/{%- endfor -%}",
            filename="{{version.nice_name}}_v{{"%03d"|format(version.version_number)}}"
        )

        s1.templates.append(f1)
        p1.structure = s1

        # now because we have defined a FilenameTemplate for Assets,
        # Stalker is now able to produce a path and a filename for any Version
        # related to an asset in this project.

    Args:
        target_entity_type (str): The class name that this FilenameTemplate is designed
            for. You can also pass the class itself. So both of the examples below can
            work::

                new_filename_template1 = FilenameTemplate(target_entity_type="Asset")
                new_filename_template2 = FilenameTemplate(target_entity_type=Asset)

            A TypeError will be raised when it is skipped or it is None and a
            ValueError will be raised when it is given as and empty string.

        path (str): A `Jinja2`_ template code which specifies the path of the given
            item. It is relative to the repository root. A typical example  could be::

            '$REPO{{project.repository.id}}/{{project.code}}/{%- for parent_task in parent_tasks -%}{{parent_task.nice_name}}/{%- endfor -%}'

        filename (str): A `Jinja2`_ template code which specifies the file name of the
            given item. It is relative to the :attr:`.FilenameTemplate.path`. A typical
            example could be::

            '{{version.nice_name}}_v{{"%03d"|format(version.version_number)}}'

            Could be set to an empty string or None, the default value is None. It can
            be None, or an empty string, or it can be skipped.

    .. _Jinja2: http://jinja.pocoo.org/docs/
    """  # noqa: B950

    __auto_name__ = False
    __strictly_typed__ = False
    __tablename__ = "FilenameTemplates"
    __mapper_args__ = {"polymorphic_identity": "FilenameTemplate"}
    filenameTemplate_id: Mapped[int] = mapped_column(
        "id", ForeignKey("Entities.id"), primary_key=True
    )

    path: Mapped[Optional[str]] = mapped_column(
        Text, doc="""The template code for the path of this FilenameTemplate."""
    )

    filename: Mapped[Optional[str]] = mapped_column(
        Text, doc="""The template code for the file part of the FilenameTemplate."""
    )

    def __init__(
        self,
        target_entity_type: Optional[str] = None,
        path: Optional[str] = None,
        filename: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        super(FilenameTemplate, self).__init__(**kwargs)
        TargetEntityTypeMixin.__init__(self, target_entity_type, **kwargs)
        self.path = path
        self.filename = filename

    @validates("path")
    def _validate_path(self, key: str, path: Union[None, str]) -> str:
        """Validate the given path value.

        Args:
            key (str): The name of the validated column.
            path (Union[None, str]): The path value to be validated.

        Raises:
            TypeError: If the given path value is not None and not a string.

        Returns:
            str: The validated path value.
        """
        # check if it is None
        if path is None:
            path = ""

        if not isinstance(path, str):
            raise TypeError(
                "{}.path attribute should be string, not {}: '{}'".format(
                    self.__class__.__name__, path.__class__.__name__, path
                )
            )

        return path

    @validates("filename")
    def _validate_filename(self, key: str, filename: Union[None, str]) -> str:
        """Validate the given filename value.

        Args:
            key (str): The name of the validated column.
            filename (Union[None, str]): The filename value to be validated.

        Raises:
            TypeError: If the given filename value is not None and not a string.

        Returns:
            str: The validated filename value.
        """
        # check if it is None
        if filename is None:
            filename = ""

        if not isinstance(filename, str):
            raise TypeError(
                "{}.filename attribute should be string, not {}: '{}'".format(
                    self.__class__.__name__, filename.__class__.__name__, filename
                )
            )

        return filename

    def __eq__(self, other: Any) -> bool:
        """Check the equality.

        Args:
            other (Any): The other object.

        Returns:
            bool: True if the other object is a FilenameTemplate instance and has the
                same target_entity_type, path and filename.
        """
        return (
            super(FilenameTemplate, self).__eq__(other)
            and isinstance(other, FilenameTemplate)
            and self.target_entity_type == other.target_entity_type
            and self.path == other.path
            and self.filename == other.filename
        )

    def __hash__(self) -> int:
        """Return the hash value of this instance.

        Because the __eq__ is overridden the __hash__ also needs to be overridden.

        Returns:
            int: The hash value.
        """
        return super(FilenameTemplate, self).__hash__()



# -*- coding: utf-8 -*-
"""Scheduler related function and classes are situated here."""

import csv
import datetime
import json
import os
import subprocess
import sys
import tempfile
import time
from typing import List, Optional, TYPE_CHECKING, Union

from jinja2 import Template

import pytz

from sqlalchemy import bindparam, text

from stalker import defaults
from stalker.db.session import DBSession
from stalker.log import get_logger
from stalker.models.project import Project
from stalker.models.task import Task, Task_Computed_Resources

if TYPE_CHECKING:  # pragma: no cover
    from stalker.models.studio import Studio

logger = get_logger(__name__)


class SchedulerBase(object):
    """This is the base class for schedulers.

    All the schedulers should be derived from this class.
    """

    def __init__(self, studio: Optional["Studio"] = None) -> None:
        self._studio = None
        self.studio = studio

    def _validate_studio(self, studio: Union[None, "Studio"]) -> Union[None, "Studio"]:
        """Validate the given studio value.

        Args:
            studio (Union[None, Studio]): The Studio instance to set the studio
                attribute to.

        Raises:
            TypeError: If the given value is not a Studio instance.

        Returns:
            Union[None, Studio]: The validated Studio instance.
        """
        if studio is not None:
            from stalker.models.studio import Studio

            if not isinstance(studio, Studio):
                raise TypeError(
                    f"{self.__class__.__name__}.studio should be an instance of "
                    "stalker.models.studio.Studio, "
                    f"not {studio.__class__.__name__}: '{studio}'"
                )
        return studio

    @property
    def studio(self) -> Union[None, "Studio"]:
        """Return studio attribute value.

        Returns:
            Studio: The studio attribute value.
        """
        return self._studio

    @studio.setter
    def studio(self, studio: Union[None, "Studio"]) -> None:
        """Set studio attribute.

        Args:
            studio (Studio): The Studio instance to set the studio attribute to.
        """
        self._studio = self._validate_studio(studio)

    def schedule(self) -> None:
        """Schedule function that needs to be implemented in the derivatives.

        Raises:
            NotImplementedError: If this is not implemented in the derived class.
        """
        raise NotImplementedError


class TaskJugglerScheduler(SchedulerBase):
    """This is the main scheduler for Stalker right now.

    This class prepares the data for TaskJuggler and let it solve the
    scheduling problem, and then retrieves the solved date and resource data
    back.

    TaskJugglerScheduler needs a :class:`.Studio` instance to work with, it
    will create a .tjp file and then solve the tasks and restore the
    computed_start and computed_end dates and the computed_resources
    attributes for each task.

    Stalker will pass all its data to TaskJuggler by creating a tjp file that
    TaskJuggler can parse. This tjp file has all the Projects, Tasks, Users,
    Departments, TimeLogs, Vacations and everything that TJ need for solving
    the tasks. With every new version of it, Stalker tries to cover more and
    more TaskJuggler directives.

    .. note::
       .. versionadded:: 0.2.5
          Alternative Resources

       Stalker is now able to pass alternative resources to TaskJuggler.
       Although, per resource alternatives are not yet possible, it will be
       implemented in future versions of Stalker.

    .. note::
       .. versionadded:: 0.2.5
          Task Dependency Relation Attributes

       Stalker now can use 'gapduration', 'gaplength', 'onstart' and 'onend'
       TaskJuggler directives for each dependent task of a task. Use the
       TaskDependency instance in Task.task_dependency attribute to control how
       a particular task is depending on another task.

    .. warning::
       **Task.computed_resources Attribute Content**

       After the scheduling is finished, TaskJuggler will create a ``csv``
       report that TaskJugglerScheduler will parse. This csv file contains the
       ``id``, ``start date``, ``end date`` and ``resources`` data. The
       resources reported back by TJ will be stored in
       :attr:`.Task.computed_resources` attribute.

       TaskJuggler will put all the resources who may have entered a
       :class:`.TimeLog` previously to the csv file. But the resources from the
       csv file may not be in :attr:`.Task.resources` or
       :attr:`.Task.alternative_resources` anymore. Because of that,
       TaskJugglerScheduler will only store the resources those are both in csv
       file and in :attr:`.Task.resources` or
       :attr:`.Task.alternative_resources` attributes.

    Stalker will export each Project to tjp as the highest task in the
    hierarchy and all the projects will be combined in to the same tjp file.
    Combining all the Projects in one tjp file has a very nice side effect,
    projects using the same resources will respect their allocations to the
    resource. So that when a TaskJugglerScheduler instance is used to schedule
    the project, all projects are scheduled together.

    The following table shows which Stalker data type is converted to which
    TaskJuggler type:

      +------------+-------------+
      | Stalker    | TaskJuggler |
      +============+=============+
      | Studio     | Project     |
      +------------+-------------+
      | Project    | Task        |
      +------------+-------------+
      | Task       | Task        |
      +------------+-------------+
      | Asset      | Task        |
      +------------+-------------+
      | Shot       | Task        |
      +------------+-------------+
      | Sequence   | Task        |
      +------------+-------------+
      | Departmemt | Resource    |
      +------------+-------------+
      | User       | Resource    |
      +------------+-------------+
      | TimeLog    | Booking     |
      +------------+-------------+
      | Vacation   | Vacation    |
      +------------+-------------+

    Args:
        compute_resources (bool): When set to True it will also consider
            :attr:`.Task.alternative_resources` attribute and will fill
            :attr:`.Task.computed_resources` attribute for each Task. With
            :class:`.TaskJugglerScheduler` when the total number of Task is around
            15k it will take around 7 minutes to generate this data, so by default
            it is False.
        parsing_method (int): Choose between SQL (0) or Pure Python (1) parsing.
            The default is SQL.
    """

    def __init__(
        self,
        studio: Optional["Studio"] = None,
        compute_resources: Optional[bool] = False,
        parsing_method: Optional[int] = 0,
        projects: Optional[Project] = None,
    ) -> None:
        super(TaskJugglerScheduler, self).__init__(studio)

        self.tjp_content = ""

        self.temp_file_full_path = None
        self.temp_file_path = None
        self.temp_file_name = None

        self.tjp_file_full_path = None
        self.tjp_file = None

        self.csv_file_full_path = None
        self.csv_file = None

        self.compute_resources = compute_resources
        self.parsing_method = parsing_method

        self._projects = []
        self.projects = projects

    def _create_tjp_file(self) -> None:
        """Create the tjp file."""
        self.temp_file_full_path = tempfile.mktemp(prefix="Stalker_")
        self.temp_file_path = os.path.dirname(self.temp_file_full_path)
        self.temp_file_name = os.path.basename(self.temp_file_full_path)
        self.tjp_file_full_path = f"{self.temp_file_full_path}.tjp"
        self.csv_file_full_path = f"{self.temp_file_full_path}.csv"

    def _create_tjp_file_content(self) -> None:  # noqa: C901
        """Create the tjp file content."""
        start = time.time()

        # use new way of doing it, it will just work with PostgreSQL
        template = Template(defaults.tjp_main_template2)

        if not self.projects:
            project_ids = (
                DBSession.connection()
                .execute(text('select id, code from "Projects"'))
                .fetchall()
            )
        else:
            project_ids = [[project.id] for project in self.projects]

        sql_query = """select
    "Tasks".id,
    tasks.path,
    coalesce("Tasks".parent_id, "Tasks".project_id) as parent_id,
    tasks.entity_type,
    tasks.name,
    "Tasks".priority,
    "Tasks".schedule_timing,
    "Tasks".schedule_unit,
    "Tasks".schedule_model,
    "Tasks".allocation_strategy,
    "Tasks".persistent_allocation,
    tasks.depth,
    task_resources.resource_ids,
    task_alternative_resources.resource_ids as alternative_resource_ids,
    time_logs.time_log_array,
    task_dependencies.dependency_info,
    not exists (
       select 1
        from "Tasks" as "Child_Tasks"
        where "Child_Tasks".parent_id = "Tasks".id
    ) as is_leaf
from "Tasks"
join (
    with recursive recursive_task(id, parent_id, path_as_text, path, depth) as (
        select
            id,
            parent_id,
            id::text as path_as_text,
            array[project_id] as path,
            0
        from "Tasks"
        where parent_id is NULL and project_id = :id
    union all
        select
            task.id,
            task.parent_id,
            (parent.path_as_text || '-' || task.id) as path_as_text,
            (parent.path || task.parent_id) as path,
            parent.depth + 1 as depth
        from "Tasks" as task
        join recursive_task as parent on task.parent_id = parent.id
    ) select
        recursive_task.id,
        recursive_task.parent_id,
        recursive_task.path_as_text,
        recursive_task.path,
        "SimpleEntities".name as name,
        "SimpleEntities".entity_type,
        recursive_task.depth
    from recursive_task
    join "SimpleEntities" on recursive_task.id = "SimpleEntities".id
    --order by path_as_text
) as tasks on "Tasks".id = tasks.id

-- resources
left outer join (
    select
        task_id,
        array_agg(resource_id order by resource_id) as resource_ids
    from "Task_Resources"
    group by task_id
) as task_resources on "Tasks".id = task_resources.task_id

-- alternative resources
left outer join (
    select
        task_id,
        array_agg(resource_id order by resource_id) as resource_ids
    from "Task_Alternative_Resources"
    group by task_id
) as task_alternative_resources on "Tasks".id = task_alternative_resources.task_id

-- time logs
left outer join (
    select
        "TimeLogs".task_id,
        array_agg(('User_' || "TimeLogs".resource_id, to_char(cast("TimeLogs".start at time zone 'utc' as timestamp), 'YYYY-MM-DD-HH24:MI:00'), to_char(cast("TimeLogs".end at time zone 'utc' as timestamp), 'YYYY-MM-DD-HH24:MI:00'))) as time_log_array
    from "TimeLogs"
    group by task_id
) as time_logs on "Tasks".id = time_logs.task_id

-- dependencies
left outer join (
    select
        task_id,
        array_agg((tasks.alt_path, dependency_target, gap_timing, gap_unit, gap_model)) dependency_info
    from "Task_Dependencies"
    join (
        with recursive recursive_task(id, parent_id, alt_path) as (
            select
                id,
                parent_id,
                project_id::text as alt_path
            from "Tasks"
            where parent_id is NULL
        union all
            select
                task.id,
                task.parent_id,
                (parent.alt_path || '-' || task.parent_id) as alt_path
            from "Tasks" as task
            join recursive_task as parent on task.parent_id = parent.id
        ) select
            recursive_task.id,
            recursive_task.parent_id,
            recursive_task.alt_path || '-' || recursive_task.id as alt_path
        from recursive_task
        join "SimpleEntities" on recursive_task.id = "SimpleEntities".id
    ) as tasks on "Task_Dependencies".depends_on_id = tasks.id
    group by task_id
) as task_dependencies on "Tasks".id = task_dependencies.task_id

--order by "Tasks".id
order by path_as_text"""  # noqa: B950

        result_buffer = []
        num_of_records = 0

        # run it per project
        for pr in project_ids:
            p_id = pr[0]
            # p_code = pr[1]

            result = DBSession.connection().execute(text(sql_query), {"id": p_id})

            # start by adding the project first
            result_buffer.append(f'task Project_{p_id} "Project_{p_id}" {{')

            # now start jumping around
            previous_level = 0
            for r in result.fetchall():
                # start by appending task tjp id first
                task_id = r[0]
                # path = r[1]
                # parent_id = r[2]
                # entity_type = r[3]
                # name = r[4]
                priority = r[5]
                schedule_timing = r[6]
                schedule_unit = r[7]
                schedule_model = r[8]
                allocation_strategy = r[9]
                persistent_allocation = r[10]
                depth = r[11] + 1
                resource_ids = r[12]
                alternative_resource_ids = r[13]
                time_log_array = r[14]
                dependency_info = r[15]
                is_leaf = r[16]

                tab = "  " * depth

                # close the previous level if necessary
                for i in range(previous_level - depth + 1):
                    i_tab = "  " * (previous_level - i)
                    result_buffer.append(f"{i_tab}}}")

                result_buffer.append(
                    f"""{tab}task Task_{task_id} "Task_{task_id}" {{"""
                )

                # append priority if it is different then 500
                if priority != 500:
                    result_buffer.append(f"{tab}  priority {priority}")

                # append dependency information
                if dependency_info:
                    dep_buffer = [f"{tab}  depends "]

                    json_data = json.loads(
                        dependency_info.replace("{", "[")
                        .replace("}", "]")
                        .replace("(", "")
                        .replace(")", "")
                    )  # it is an array of string

                    for i, dependency in enumerate(json_data):
                        if i > 0:
                            dep_buffer.append(", ")

                        (
                            dep_full_ids,
                            dependency_target,
                            gap_timing,
                            gap_unit,
                            gap_model,
                        ) = dependency.split(",")

                        dep_full_path = ".".join(
                            map(lambda x: f"Task_{x}", dep_full_ids.split("-"))
                        )
                        # fix for Project id
                        dep_full_path = f"Project_{dep_full_path[5:]}"

                        dep_string = f"{dep_full_path} {{{dependency_target}}}"

                        dep_buffer.append(dep_string)

                    result_buffer.append("".join(dep_buffer))

                # append schedule model and timing information
                # if this is a leaf task and has resources
                if is_leaf and resource_ids:
                    result_buffer.append(
                        f"{tab}  {schedule_model} {schedule_timing}{schedule_unit}"
                    )

                    resource_buffer = [f"{tab}  allocate "]
                    for i, resource_id in enumerate(resource_ids):
                        if i > 0:
                            resource_buffer.append(", ")
                        resource_buffer.append(f"User_{resource_id}")

                        # now go through alternatives
                        if alternative_resource_ids:
                            resource_buffer.append(" { alternative ")
                            for j, alt_resource_id in enumerate(
                                alternative_resource_ids
                            ):
                                if j > 0:
                                    resource_buffer.append(", ")
                                resource_buffer.append(f"User_{alt_resource_id}")

                            # set the allocation strategy
                            resource_buffer.append(f" select {allocation_strategy}")

                            # is is persistent
                            if persistent_allocation:
                                resource_buffer.append(" persistent")
                            resource_buffer.append(" }")

                    result_buffer.append("".join(resource_buffer))

                    # append any time log information
                    if time_log_array:
                        json_data = json.loads(
                            time_log_array.replace("{", "[")
                            .replace("}", "]")
                            .replace("(", "")
                            .replace(")", "")
                        )  # it is an array of string

                        for tlog in json_data:
                            user_id, t_start, t_end = tlog.split(",")
                            result_buffer.append(
                                f"{tab}  booking {user_id} {t_start} - {t_end} "
                                "{ overtime 2 }"
                            )

                previous_level = depth
                num_of_records += 1

            # and close the brackets per project
            depth = 0  # current depth is 0 (Project)
            # previous_level is the last task
            for i in range(previous_level - depth + 1):
                i_tab = "  " * (previous_level - i)
                result_buffer.append(f"{i_tab}}}")

        tasks_buffer = "\n".join(result_buffer)

        import stalker

        self.tjp_content = template.render(
            {
                "stalker": stalker,
                "studio": self.studio,
                "csv_file_name": self.temp_file_name,
                "csv_file_full_path": self.temp_file_full_path,
                "compute_resources": self.compute_resources,
                "tasks_buffer": tasks_buffer,
            },
            trim_blocks=True,
            lstrip_blocks=True,
        )

        logger.debug(f"total number of records: {num_of_records}")

        end = time.time()
        logger.debug(
            "rendering the whole tjp file took: {:0.3f} seconds".format(end - start)
        )

    def _fill_tjp_file(self) -> None:
        """Fill the tjp file with content."""
        with open(self.tjp_file_full_path, "w+") as self.tjp_file:
            self.tjp_file.write(self.tjp_content)

    def _delete_tjp_file(self) -> None:
        """Delete the temp tjp file."""
        try:
            os.remove(self.tjp_file_full_path)
        except OSError:
            pass

    def _delete_csv_file(self) -> None:
        """Delete the temp csv file."""
        try:
            os.remove(self.csv_file_full_path)
        except OSError:
            pass

    def _clean_up(self) -> None:
        """Remove the temp files."""
        self._delete_tjp_file()
        self._delete_csv_file()

    def _parse_csv_file(self) -> None:
        """Parse the csv file and set the Task.computes_start and Task.computed_end."""
        parsing_start = time.time()

        logger.debug(f"csv_file_full_path : {self.csv_file_full_path}")
        if not os.path.exists(self.csv_file_full_path):
            logger.debug("could not find CSV file, returning without updating db!")
            return

        entity_ids = []
        update_data = []
        update_user_data = []

        with open(self.csv_file_full_path, "r") as self.csv_file:
            csv_content = csv.reader(self.csv_file, delimiter=";")

            lines = [line for line in csv_content]
            lines.pop(0)

        for data in lines:
            id_line = data[0]

            entity_id = int(id_line.split(".")[-1].split("_")[-1])
            if not entity_id:
                continue

            entity_ids.append(entity_id)
            start_date = datetime.datetime.strptime(data[1], "%Y-%m-%d-%H:%M")
            end_date = datetime.datetime.strptime(data[2], "%Y-%m-%d-%H:%M")

            # implement time zone info
            start_date = start_date.replace(tzinfo=pytz.utc)
            end_date = end_date.replace(tzinfo=pytz.utc)

            # computed_resources
            if self.compute_resources and data[3] != "":
                resources_data = map(
                    lambda x: x.split("_")[-1].split(")")[0], data[3].split(",")
                )
                for rid in resources_data:
                    update_user_data.append({"task_id": entity_id, "resource_id": rid})

            update_data.append(
                {
                    "b_id": entity_id,
                    "start": start_date,
                    "end": end_date,
                    "computed_start": start_date,
                    "computed_end": end_date,
                }
            )

        # update date values
        update_statement = (
            Task.__table__.update()
            .where(Task.__table__.c.id == bindparam("b_id"))
            .values(
                start=bindparam("start"),
                end=bindparam("end"),
                computed_start=bindparam("computed_start"),
                computed_end=bindparam("computed_end"),
            )
        )
        DBSession.connection().execute(update_statement, update_data)

        # update project dates
        update_project_statement = (
            Project.__table__.update()
            .where(Project.__table__.c.id == bindparam("b_id"))
            .values(
                start=bindparam("start"),
                end=bindparam("end"),
                computed_start=bindparam("computed_start"),
                computed_end=bindparam("computed_end"),
            )
        )
        DBSession.connection().execute(update_project_statement, update_data)

        # update computed resources data
        # first delete everything
        if self.compute_resources:
            delete_resources_statement = Task_Computed_Resources.delete()

            update_resources_statement = Task_Computed_Resources.insert().values(
                task_id=bindparam("task_id"), resource_id=bindparam("resource_id")
            )

            DBSession.connection().execute(delete_resources_statement)
            DBSession.connection().execute(update_resources_statement, update_user_data)

        parsing_end = time.time()
        logger.debug(
            "completed parsing csv file in (SQL): {} seconds".format(
                parsing_end - parsing_start
            )
        )

    def schedule(self) -> str:
        """Schedule the project or all projects in the Studio.

        Raises:
            TypeError: If the self.studio is not a Studio instance.
            RuntimeError: If the tj3 command returns an error.

        Returns:
            str: The tj3 command output.
        """
        # check the studio attribute
        from stalker.models.studio import Studio

        if not isinstance(self.studio, Studio):
            raise TypeError(
                f"{self.__class__.__name__}.studio should be an instance of "
                "stalker.models.studio.Studio, "
                f"not {self.studio.__class__.__name__}: '{self.studio}'"
            )

        # create a tjp file
        self._create_tjp_file()

        # create tjp file content
        self._create_tjp_file_content()

        # fill it with data
        self._fill_tjp_file()

        logger.debug(f"tjp_file_full_path: {self.tjp_file_full_path}")

        # pass it to tj3
        if sys.platform == "win32":
            logger.debug("tj3 using fallback mode for Windows!")
            command = "{} {} -o {}".format(
                defaults.tj_command,
                self.tjp_file_full_path,
                self.temp_file_path,
            )
            logger.debug(f"tj3 command: {command}")
            return_code = os.system(command)
            stderr_buffer = ""
        else:
            process = subprocess.Popen(
                [
                    defaults.tj_command,
                    self.tjp_file_full_path,
                    "-o",
                    self.temp_file_path,
                ],
                stderr=subprocess.PIPE,
            )

            # loop until process finishes and capture stderr output
            stderr_buffer = []
            while True:
                stderr = process.stderr.readline()

                if stderr == b"" and process.poll() is not None:
                    break

                if stderr != b"":
                    stderr = stderr.decode("utf-8").strip()
                    stderr_buffer.append(stderr)
                    logger.debug(stderr)

            # flatten the buffer
            stderr_buffer = "\n".join(stderr_buffer)

            return_code = process.returncode

        if return_code:
            # there is an error
            raise RuntimeError(stderr_buffer)

        # read back the csv file
        self._parse_csv_file()

        logger.debug(f"tj3 return code: {return_code}")

        # remove the tjp file
        self._clean_up()

        return stderr_buffer

    def _validate_projects(self, projects: List[Project]) -> List[Project]:
        """Validate the given projects value.

        Args:
            projects (List[Project]): List of Project instances.

        Raises:
            TypeError: If the projects is not a list or if any of the items in the
                projects list is not a Project instance.

        Returns:
            List[Project]: List of validated Project instances.
        """
        if projects is None:
            projects = []

        msg = (
            "{cls}.projects should only contain instances of "
            "stalker.models.project.Project, not "
            "{projects_class}: '{projects}'"
        )

        if not isinstance(projects, list):
            raise TypeError(
                msg.format(
                    cls=self.__class__.__name__,
                    projects_class=projects.__class__.__name__,
                    projects=projects,
                )
            )

        for item in projects:
            if not isinstance(item, Project):
                raise TypeError(
                    msg.format(
                        cls=self.__class__.__name__,
                        projects_class=item.__class__.__name__,
                        projects=item,
                    )
                )

        return projects

    @property
    def projects(self) -> List[Project]:
        """Return the projects attribute value.

        Returns:
            List[Project]: List of Project instances.
        """
        return self._projects

    @projects.setter
    def projects(self, projects: List[Project]) -> None:
        """Set the projects attribute.

        Args:
            projects (List[Project]): List of Project instances.
        """
        self._projects = self._validate_projects(projects)



# -*- coding: utf-8 -*-
"""stalker.testing module."""

import pytest
from tests.utils import get_server_details_from_url


@pytest.mark.parametrize(
    "url,expected",
    [
        (
            "postgresql://postgres:postgres@localhost:5432/stalker_test_e0b9bc6a",
            {
                "dialect": "postgresql",
                "username": "postgres",
                "password": "postgres",
                "hostname": "localhost",
                "port": "5432",
                "database_name": "stalker_test_e0b9bc6a",
            },
        ),
        (
            "postgresql://postgres:postgres@localhost/stalker_test_e0b9bc6a",
            {
                "dialect": "postgresql",
                "username": "postgres",
                "password": "postgres",
                "hostname": "localhost",
                "port": "",
                "database_name": "stalker_test_e0b9bc6a",
            },
        ),
    ],
)
def test_get_server_details_from_url(url, expected):
    """get_server_details_from_url."""
    assert get_server_details_from_url(url) == expected



# -*- coding: utf-8 -*-
"""Mixins are situated here."""

import datetime
from typing import (
    Any,
    Dict,
    Generator,
    List,
    Optional,
    TYPE_CHECKING,
    Tuple,
    Type,
    Union,
)
from typing_extensions import Self

import pytz

from sqlalchemy import (
    Column,
    Float,
    ForeignKey,
    Integer,
    Interval,
    String,
    Table,
)
from sqlalchemy.exc import OperationalError, UnboundExecutionError
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.orm import (
    Mapped,
    backref,
    mapped_column,
    relationship,
    synonym,
    validates,
)

from stalker import defaults
from stalker.db.declarative import Base
from stalker.db.session import DBSession
from stalker.db.types import GenericDateTime
from stalker.log import get_logger
from stalker.models.enum import (
    ScheduleConstraint,
    ScheduleConstraintDecorator,
    ScheduleModel,
    ScheduleModelDecorator,
    TimeUnit,
    TimeUnitDecorator,
    TraversalDirection,
)
from stalker.utils import check_circular_dependency, make_plural, walk_hierarchy


if TYPE_CHECKING:  # pragma: no cover
    from stalker.models.auth import Permission
    from stalker.models.project import Project
    from stalker.models.status import Status, StatusList
    from stalker.models.file import File
    from stalker.models.studio import WorkingHours


logger = get_logger(__name__)


def create_secondary_table(
    primary_cls_name: str,
    secondary_cls_name: str,
    primary_cls_table_name: str,
    secondary_cls_table_name: str,
    secondary_table_name: Optional[str] = None,
) -> Table:
    """Create any secondary table.

    Args:
        primary_cls_name (str): The primary class name.
        secondary_cls_name (str): The secondary class name.
        primary_cls_table_name (str): The primary class table name.
        secondary_cls_table_name (str): The secondary class table name.
        secondary_table_name (Union[None, str]): Optional secondary table name.

    Raises:
        TypeError: If primary_cls_name is not a str.
        TypeError: If secondary_cls_name is not a str.
        TypeError: If primary_cls_table_name is not a str.
        TypeError: If secondary_cls_table_name is not a str.
        TypeError: If secondary_table_name is not a str.
        ValueError: If primary_cls_name is an empty str.
        ValueError: If secondary_cls_name is an empty str.
        ValueError: If primary_cls_table_name is an empty str.
        ValueError: If secondary_cls_table_name is an empty str.
        ValueError: If secondary_table_name is an empty str.

    Returns:
        Table: The secondary table.
    """
    # validate data
    # primary_cls_name
    if not isinstance(primary_cls_name, str):
        raise TypeError(
            "primary_cls_name should be a str containing the primary class name, "
            f"not {primary_cls_name.__class__.__name__}: '{primary_cls_name}'"
        )

    if primary_cls_name == "":
        raise ValueError(
            "primary_cls_name should be a str containing the primary class name, "
            f"not: '{primary_cls_name}'"
        )

    # secondary_cls_name
    if not isinstance(secondary_cls_name, str):
        raise TypeError(
            "secondary_cls_name should be a str containing the secondary class name, "
            f"not {secondary_cls_name.__class__.__name__}: '{secondary_cls_name}'"
        )

    if secondary_cls_name == "":
        raise ValueError(
            "secondary_cls_name should be a str containing the secondary class name, "
            f"not: '{secondary_cls_name}'"
        )

    # primary_cls_table_name
    if not isinstance(primary_cls_table_name, str):
        raise TypeError(
            "primary_cls_table_name should be a str containing the primary class "
            f"table name, not {primary_cls_table_name.__class__.__name__}: "
            f"'{primary_cls_table_name}'"
        )

    if primary_cls_table_name == "":
        raise ValueError(
            "primary_cls_table_name should be a str containing the primary class "
            f"table name, not: '{primary_cls_table_name}'"
        )

    # secondary_cls_table_name
    if not isinstance(secondary_cls_table_name, str):
        raise TypeError(
            "secondary_cls_table_name should be a str containing the secondary class "
            f"table name, not {secondary_cls_table_name.__class__.__name__}: "
            f"'{secondary_cls_table_name}'"
        )

    if secondary_cls_table_name == "":
        raise ValueError(
            "secondary_cls_table_name should be a str containing the secondary class "
            f"table name, not: '{secondary_cls_table_name}'"
        )

    # secondary_table_name
    if secondary_table_name is not None and not isinstance(secondary_table_name, str):
        raise TypeError(
            "secondary_table_name should be a str containing the secondary table "
            "name, or it can be None or an empty string to let Stalker to auto "
            f"generate one, not {secondary_table_name.__class__.__name__}: "
            f"'{secondary_table_name}'"
        )

    plural_secondary_cls_name = make_plural(secondary_cls_name)

    # use the given class_name and the class_table
    if not secondary_table_name:
        secondary_table_name = f"{primary_cls_name}_{plural_secondary_cls_name}"

    # check if the table is already defined
    if secondary_table_name not in Base.metadata:
        secondary_table = Table(
            secondary_table_name,
            Base.metadata,
            Column(
                f"{primary_cls_name.lower()}_id",
                Integer,
                ForeignKey(f"{primary_cls_table_name}.id"),
                primary_key=True,
            ),
            Column(
                f"{secondary_cls_name.lower()}_id",
                Integer,
                ForeignKey(f"{secondary_cls_table_name}.id"),
                primary_key=True,
            ),
        )
    else:
        secondary_table = Base.metadata.tables[secondary_table_name]

    return secondary_table


class TargetEntityTypeMixin(object):
    """Adds target_entity_type attribute to mixed in class.

    Args:
        target_entity_type (Union[str, type]): The target entity type which this class
            is designed for. Should be a class or a class name.

        For example::

            from stalker import SimpleEntity, TargetEntityTypeMixin, Project

            class A(SimpleEntity, TargetEntityTypeMixin):
                __tablename__ = "As"
                __mapper_args__ = {"polymorphic_identity": "A"}

                def __init__(self, **kwargs):
                    super(A, self).__init__(**kwargs)
                    TargetEntityTypeMixin.__init__(self, **kwargs)

            a_obj = A(target_entity_type=Project)

        The ``a_obj`` will only be accepted by :class:`.Project` instances. You cannot
        assign it to any other class which accepts a :class:`.Type` instance.

        To control the mixed-in class behavior add these class variables to the
        mixed in class:

            __nullable_target__ : controls if the target_entity_type can be
                nullable or not. Default is False.

            __unique_target__ : controls if the target_entity_type should be unique, so
                there is only one object for one type. Default is False.
    """

    __nullable_target__ = False
    __unique_target__ = False

    @declared_attr
    def _target_entity_type(cls) -> Mapped[str]:
        """Create the _target_entity_type attribute as a declared attribute.

        Returns:
            Column: The Column related to the _target_entity_type attribute.
        """
        return mapped_column(
            "target_entity_type",
            String(128),
            nullable=cls.__nullable_target__,
            unique=cls.__unique_target__,
        )

    def __init__(self, target_entity_type: Optional[str] = None, **kwargs) -> None:
        self._target_entity_type = self._validate_target_entity_type(target_entity_type)

    def _validate_target_entity_type(self, target_entity_type: Union[str, Type]) -> str:
        """Validate the given target_entity_type value.

        Args:
            target_entity_type (Union[str, type]): The target_entity_type that this
                entity is valid for.

        Raises:
            TypeError: If the given target_entity_type value is None.
            ValueError: If the given target_entity_type value is an empty str.

        Returns:
            str: The validated target_entity_type value.
        """
        # it cannot be None
        if target_entity_type is None:
            raise TypeError(
                f"{self.__class__.__name__}.target_entity_type cannot be None"
            )

        # check if it is a class
        if isinstance(target_entity_type, type):
            target_entity_type = target_entity_type.__name__

        if target_entity_type == "":
            raise ValueError(
                f"{self.__class__.__name__}.target_entity_type cannot be empty"
            )

        return target_entity_type

    def _target_entity_type_getter(self) -> str:
        """Return the _target_entity_type attribute value.

        Returns:
            str: The _target_entity_type attribute value.
        """
        return self._target_entity_type

    @declared_attr
    def target_entity_type(cls) -> Mapped[str]:
        """Create the target_entity_type attribute as a declared attribute.

        Returns:
            SynonymProperty: The target_entity_type property.
        """
        return synonym(
            "_target_entity_type",
            descriptor=property(
                fget=cls._target_entity_type_getter,
                doc="""The entity type which this object is valid for.

                Usually it is set to the TargetClass directly.
                """,
            ),
        )


class StatusMixin(object):
    """Makes the mixed in object statusable.

    This mixin adds status and status_list attributes to the mixed in class.
    Any object that needs a status and a corresponding status list can include
    this mixin.

    When mixed with a class which don't have an __init__ method, the mixin
    supplies one, and in this case the parameters below must be defined.

    Args:
        status_list (StatusList): this attribute holds a status list object, which
            shows the possible statuses that this entity could be in. This attribute
            cannot be empty or None. Giving a StatusList object, the
            StatusList.target_entity_type should match the current class.

            .. versionadded:: 0.1.2.a4

                The status_list argument now can be skipped or can be None if there
                is an active database connection and there is a suitable
                :class:`.StatusList` instance in the database whom
                :attr:`.StatusList.target_entity_type` attribute is set to the
                current mixed-in class name.

        status (Status): It is a :class:`.Status` instance which shows the current
            status of the statusable object. Integer values are also accepted,
            which shows the index of the desired status in the ``status_list``
            attribute of the current statusable object. If a :class:`.Status`
            instance is supplied, it should also be present in the ``status_list``
            attribute. If set to None then the first :class:`.Status` instance
            in the ``status_list`` will be used.

            .. versionadded:: 0.2.0

                Status attribute as Status instance:

                It is now possible to set the status of the instance by a
                :class:`.Status` instance directly. And the :attr:`.StatusMixin.status`
                will return a proper :class:`.Status` instance.
    """

    def __init__(
        self,
        status: Union[None, "Status"] = None,
        status_list: Union[None, "StatusList"] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        self.status_list = status_list
        self.status = status

    @declared_attr
    def status_id(cls) -> Mapped[int]:
        """Create the status_id attribute as a declared attribute.

        Returns:
            Column: The Column related to the status_id attribute.
        """
        return mapped_column(
            "status_id",
            ForeignKey("Statuses.id"),
            nullable=False,
            # This is set to nullable=True but it is impossible to set the
            # status to None by using this Declarative approach.
            #
            # This is done in that way cause SQLAlchemy was flushing the data
            # (AutoFlush) preliminarily while checking if the given Status was
            # in the related StatusList, and it was complaining about the
            # status cannot be null
        )

    @declared_attr
    def status(cls) -> Mapped["Status"]:
        """Create the status attribute as a declared attribute.

        Returns:
            relationship: The relationship object related to the status attribute.
        """
        return relationship(
            "Status",
            primaryjoin=f"{cls.__name__}.status_id==Status.status_id",
            doc="""The current status of the object.

            It is a :class:`.Status` instance which
            is one of the Statuses stored in the ``status_list`` attribute
            of this object.
            """,
        )

    @declared_attr
    def status_list_id(cls) -> Mapped[int]:
        """Create the status_list_id attribute as a declared attribute.

        Returns:
            Column: The Column related to the status_list_id attribute.
        """
        return mapped_column(
            "status_list_id", ForeignKey("StatusLists.id"), nullable=False
        )

    @declared_attr
    def status_list(cls) -> Mapped["StatusList"]:
        """Create the status_list attribute as a declared attribute.

        Returns:
            relationship: The relationship object related to the status_list attribute.
        """
        return relationship(
            "StatusList",
            primaryjoin=f"{cls.__name__}.status_list_id==StatusList.status_list_id",
        )

    @validates("status_list")
    def _validate_status_list(
        self, key: str, status_list: Union[None, "StatusList"]
    ) -> "StatusList":
        """Validate the given status_list value.

        Args:
            key (str): The name of the validated column.
            status_list (Union[None, StatusList]): The status_list value to be
                validated.

        Raises:
            TypeError: If the given status_list value is not a StatusList instance.

        Returns:
            StatusList: The validated status_list value.
        """
        from stalker.models.status import StatusList

        super_names = [mro.__name__ for mro in self.__class__.__mro__]

        if status_list is None:
            # check if there is a db setup and try to get the appropriate
            # StatusList from the database

            # disable autoflush to prevent premature class initialization
            with DBSession.no_autoflush:
                try:
                    # try to get a StatusList with the target_entity_type is
                    # matching the class name
                    status_list = StatusList.query.filter(
                        StatusList.target_entity_type.in_(super_names)
                    ).first()
                except (UnboundExecutionError, OperationalError):
                    # it is not mapped just skip it
                    pass

        # if it is still None
        if status_list is None:
            # there is no db so raise an error because there is no way
            # to get an appropriate StatusList
            raise TypeError(
                f"{self.__class__.__name__} instances cannot be initialized without a "
                "stalker.models.status.StatusList instance, please pass a "
                "suitable StatusList "
                f"(StatusList.target_entity_type={self.__class__.__name__}) with the "
                "'status_list' argument"
            )
        else:
            # it is not an instance of status_list
            if not isinstance(status_list, StatusList):
                raise TypeError(
                    f"{self.__class__.__name__}.status_list should be an instance of "
                    "stalker.models.status.StatusList, "
                    f"not {status_list.__class__.__name__}: '{status_list}'"
                )

            # check if the entity_type matches to the
            # StatusList.target_entity_type
            if status_list.target_entity_type not in super_names:
                raise TypeError(
                    "The given StatusLists' target_entity_type is "
                    f"{status_list.target_entity_type}, "
                    "whereas the entity_type of this object is "
                    f"{self.__class__.__name__}"
                )

        return status_list

    @validates("status")
    def _validate_status(self, key: str, status: "Status") -> "Status":
        """Validate the given status value.

        Args:
            key (str): The name of the validated column.
            status (Status): The status value to be validated.

        Raises:
            TypeError: If the given status value is not a Status instance or an int.
            ValueError: If the status is a negative int value or the given int value
                is equal or bigger than the length of the `StatusList.statuses` or if
                the given Status instance is not in the `StatusList.statuses` list.

        Returns:
            Status: The validated status value.
        """
        from stalker.models.status import Status

        # it is set to None
        if status is None:
            with DBSession.no_autoflush:
                status = self.status_list.statuses[0]

        # it is not an instance of status or int
        if not isinstance(status, (Status, int)):
            raise TypeError(
                f"{self.__class__.__name__}.status must be an instance of "
                "stalker.models.status.Status or an integer showing the index of the "
                f"Status object in the {self.__class__.__name__}.status_list, "
                f"not {status.__class__.__name__}: '{status}'"
            )

        if isinstance(status, int):
            # if it is not in the correct range:
            if status < 0:
                raise ValueError(
                    f"{self.__class__.__name__}.status must be a non-negative integer"
                )

            if status >= len(self.status_list.statuses):
                raise ValueError(
                    f"{self.__class__.__name__}.status cannot be bigger than the "
                    "length of the status_list"
                )
                # get the status instance out of the status_list instance
            status = self.status_list[status]

        # check if the given status is in the status_list
        if status not in self.status_list:
            raise ValueError(
                f"The given Status instance for {self.__class__.__name__}.status is "
                f"not in the {self.__class__.__name__}.status_list, please supply a "
                "status from that list."
            )

        return status


class DateRangeMixin(object):
    """Adds date range info to the mixed in class.

    Adds date range information like ``start``, ``end`` and ``duration``. These
    attributes will be used in TaskJuggler. Because ``effort`` is only
    meaningful if there are some ``resources`` this attribute has been left
    special for :class:`.Task` class. The ``length`` has not been implemented
    because of its rare use.

    The preceding order for the attributes is as follows::

      start > end > duration

    So if all of the parameters are given only the ``start`` and the ``end``
    will be used and the ``duration`` will be calculated accordingly. In any
    other conditions the missing parameter will be calculated from the
    following table:

    +-------+-----+----------+----------------------------------------+
    | start | end | duration | DEFAULTS                               |
    +=======+=====+==========+========================================+
    |       |     |          | start = datetime.datetime.now(pytz.utc)|
    |       |     |          |                                        |
    |       |     |          | duration = datetime.timedelta(days=10) |
    |       |     |          |                                        |
    |       |     |          | end = start + duration                 |
    +-------+-----+----------+----------------------------------------+
    |   X   |     |          | duration = datetime.timedelta(days=10) |
    |       |     |          |                                        |
    |       |     |          | end = start + duration                 |
    +-------+-----+----------+----------------------------------------+
    |   X   |  X  |          | duration = end - start                 |
    +-------+-----+----------+----------------------------------------+
    |   X   |     |    X     | end = start + duration                 |
    +-------+-----+----------+----------------------------------------+
    |   X   |  X  |    X     | duration = end - start                 |
    +-------+-----+----------+----------------------------------------+
    |       |  X  |    X     | start = end - duration                 |
    +-------+-----+----------+----------------------------------------+
    |       |  X  |          | duration = datetime.timedelta(days=10) |
    |       |     |          |                                        |
    |       |     |          | start = end - duration                 |
    +-------+-----+----------+----------------------------------------+
    |       |     |    X     | start = datetime.datetime.now(pytz.utc)|
    |       |     |          |                                        |
    |       |     |          | end = start + duration                 |
    +-------+-----+----------+----------------------------------------+

    Only the ``start``, ``end`` will be stored. The ``duration`` attribute is
    the direct difference of the the ``start`` and ``end`` attributes, so there
    is no need to store it. But if will be used in calculation of the start and
    end values.

    The start and end attributes have a ``computed`` companion. Which are the
    return values from TaskJuggler. so for ``start`` there is the
    ``computed_start`` and for ``end`` there is the ``computed_end``
    attributes.

    The date attributes can be managed with timezones. Follow the Python idioms
    shown in the `documentation of datetime`_

    .. _documentation of datetime: https://docs.python.org/library/datetime.html

    Args:
        start (datetime.datetime): the start date of the entity, should be a
            datetime.datetime instance, the start is the pin point for the date
            calculation. In any condition if the start is available then the value
            will be preserved. If start passes the end the end is also changed
            to a date to keep the timedelta between dates. The default value is
            datetime.datetime.now(pytz.utc)

        end (datetime.datetime): the end of the entity, should be a datetime.datetime
            instance, when the start is changed to a date passing the end, then
            the end is also changed to a later date so the timedelta between the
            dates is kept.

        duration (datetime.timedelta): The duration of the entity. It is a
            :class:`datetime.timedelta` instance. The default value is read from
            he :class:`.Config` class. See the table above for the initialization
            rules.
    """

    def __init__(
        self,
        start: Optional[datetime.datetime] = None,
        end: Optional[datetime.datetime] = None,
        duration: Optional[datetime.timedelta] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        self._start, self._end, self._duration = self._validate_dates(
            start, end, duration
        )

    @declared_attr
    def _end(cls) -> Mapped[Optional[datetime.datetime]]:
        return mapped_column("end", GenericDateTime)

    def _end_getter(self) -> datetime.datetime:
        """Return the date that the entity should be delivered.

        The end can be set to a datetime.timedelta and in this case it will be
        calculated as an offset from the start and converted to datetime.datetime again.
        Setting the start to a date passing the end will also set the end, so the
        timedelta between them is preserved, default value is 10 days.

        Returns:
            datetime.datetime: The end datetime.
        """
        with DBSession.no_autoflush:
            return self._end

    def _end_setter(self, end: datetime.datetime) -> None:
        """Set the end attribute value.

        Args:
            end (datetime.datetime): The end datetime value.
        """
        self._start, self._end, self._duration = self._validate_dates(
            self.start, end, self.duration
        )

    @declared_attr
    def end(cls) -> Mapped[Optional[datetime.datetime]]:
        """Create the end attribute as a declared attribute.

        Returns:
            SynonymProperty: The end property.
        """
        return synonym("_end", descriptor=property(cls._end_getter, cls._end_setter))

    @declared_attr
    def _start(cls) -> Mapped[Optional[datetime.datetime]]:
        """Create the start attribute as a declared attribute.

        Returns:
            Column: The Column related to the start attribute.
        """
        return mapped_column("start", GenericDateTime)

    def _start_getter(self) -> datetime.datetime:
        """Return the date that this entity should start.

        Also effects the :attr:`.DateRangeMixin.end` attribute value in certain
        conditions, if the :attr:`.DateRangeMixin.start` is set to a time passing the
        :attr:`.DateRangeMixin.end` it will also offset the :attr:`.DateRangeMixin.end`
        to keep the :attr:`.DateRangeMixin.duration` value fixed.
        :attr:`.DateRangeMixin.start` should be an instance of class:`datetime.datetime`
        and the default value is :func:`datetime.datetime.now(pytz.utc)`.

        Returns:
            datetime.datetime: The start datetime value.
        """
        with DBSession.no_autoflush:
            return self._start

    def _start_setter(self, start: datetime.datetime) -> None:
        """Set the start attribute.

        Args:
            start (datetime.datetime): The start date and time.
        """
        self._start, self._end, self._duration = self._validate_dates(
            start, self.end, self.duration
        )

    @declared_attr
    def start(cls) -> Mapped[Optional[datetime.datetime]]:
        """Create the start attribute as a declared attribute.

        Returns:
            SynonymProperty: The start property.
        """
        return synonym(
            "_start",
            descriptor=property(
                cls._start_getter,
                cls._start_setter,
            ),
        )

    @declared_attr
    def _duration(cls) -> Mapped[Optional[datetime.timedelta]]:
        """Create the duration attribute as a declared attribute.

        Returns:
            Column: The Column related to the duration attribute.
        """
        return mapped_column("duration", Interval)

    def _duration_getter(self) -> datetime.timedelta:
        """Return the duration value.

        Returns:
            datetime.timedelta: The duration value.
        """
        with DBSession.no_autoflush:
            return self._duration

    def _duration_setter(self, duration: datetime.timedelta) -> None:
        """Set the duration value.

        Args:
            duration (datetime.timedelta): The duration value.
        """
        if duration is not None:
            if isinstance(duration, datetime.timedelta):
                # set the end to None
                # to make it recalculated
                self._start, self._end, self._duration = self._validate_dates(
                    self.start, None, duration
                )
            else:
                # use the end
                self._start, self._end, self._duration = self._validate_dates(
                    self.start, self.end, duration
                )
        else:
            self._start, self._end, self._duration = self._validate_dates(
                self.start, self.end, duration
            )

    @declared_attr
    def duration(self) -> Mapped[Optional[datetime.timedelta]]:
        """Return the duration attr as a synonym.

        Returns:
            SynonymProperty: The duration property.
        """
        return synonym(
            "_duration",
            descriptor=property(
                self._duration_getter,
                self._duration_setter,
                doc="""Duration of the entity.

                It is a datetime.timedelta instance. Showing the difference of
                the :attr:`.start` and the :attr:`.end`. If edited it changes
                the :attr:`.end` attribute value.""",
            ),
        )

    def _validate_dates(
        self,
        start: datetime.datetime,
        end: datetime.datetime,
        duration: datetime.timedelta,
    ) -> Tuple[datetime.datetime, datetime.datetime, datetime.timedelta]:  # noqa: C901
        """Update the date values.

        Args:
            start (datetime.datetime): The start datetime value.
            end (datetime.datetime): The end datetime value.
            duration (datetime.timedelta): The duration value.

        Returns:
            Tuple(datetime.datetime, datetime.datetime, datetime.timedelta): The
                validated and calculated start, end dates and duration value.
        """
        # logger.debug(f"start    : {start}")
        # logger.debug(f"end      : {end}")
        # logger.debug(f"duration : {duration}")
        if not isinstance(start, datetime.datetime):
            start = None

        if not isinstance(end, datetime.datetime):
            end = None

        if not isinstance(duration, datetime.timedelta):
            duration = None

        # check start
        if start is None:
            # try to calculate the start from end and duration
            if end is None:
                # set the defaults
                start = datetime.datetime.now(pytz.utc)

                if duration is None:
                    # set the defaults
                    duration = defaults.timing_resolution

                end = start + duration
            else:
                if duration is None:
                    duration = defaults.timing_resolution

                # try:
                start = end - duration
                # except OverflowError: # end is datetime.datetime.min
                #     start = end

        # check end
        if end is None:
            if duration is None:
                duration = defaults.timing_resolution

            end = start + duration

        if end < start:
            # check duration
            if duration is None or duration < datetime.timedelta(1):
                duration = datetime.timedelta(1)

            # try:
            end = start + duration
            # except OverflowError: # start is datetime.datetime.max
            #     end = start

        # round the dates to the timing_resolution
        rounded_start = self.round_time(start)
        rounded_end = self.round_time(end)
        rounded_duration = rounded_end - rounded_start

        if rounded_duration < defaults.timing_resolution:
            rounded_duration = defaults.timing_resolution
            rounded_end = rounded_start + rounded_duration

        return rounded_start, rounded_end, rounded_duration

    @declared_attr
    def computed_start(cls) -> Mapped[Optional[datetime.datetime]]:
        """Create the computed_start attribute as a declared attribute.

        Returns:
            Column: The Column related to the computed_start attribute.
        """
        return mapped_column("computed_start", GenericDateTime)

    @declared_attr
    def computed_end(cls) -> Mapped[Optional[datetime.datetime]]:
        """Create the computed_end attribute as a declared attribute.

        Returns:
            Column: The Column related to the computed_end attribute.
        """
        return mapped_column("computed_end", GenericDateTime)

    @property
    def computed_duration(self) -> datetime.timedelta:
        """Calculate the computed duration.

        The computed_duration is calculated as the difference of computed_start and
        computed_end if there are computed_start and computed_end otherwise returns
        None.

        Returns:
            Union[None, datetime.timedelta]: None if one of computed_start or
                computed_end value is None else the difference as datetime.timedelta
                instance.
        """
        return (
            self.computed_end - self.computed_start
            if self.computed_end and self.computed_start
            else None
        )

    @classmethod
    def round_time(cls, dt: datetime.datetime) -> datetime.datetime:
        """Round the given datetime object to the defaults.timing_resolution.

        Use the  :class:`stalker.defaults.timing_resolution` as the closest number of
        seconds to round to.

        Based on Thierry Husson's answer in `Stackoverflow`_

        _`Stackoverflow` : https://stackoverflow.com/a/10854034/1431079

        Args:
            dt (datetime.datetime): The datetime object, defaults to now.

        Returns:
            datetime.datetime: The rounded datetime.datetime instance.
        """
        # to be compatible with python 2.6 use the following instead of
        # total_seconds()
        timing_resolution = defaults.timing_resolution
        trs = timing_resolution.days * 86400 + timing_resolution.seconds

        # convert to seconds
        epoch = datetime.datetime(1970, 1, 1, tzinfo=pytz.utc)

        diff = dt - epoch
        diff_in_seconds = diff.days * 86400 + diff.seconds
        return epoch + datetime.timedelta(
            seconds=(diff_in_seconds + trs * 0.5) // trs * trs
        )

    @property
    def total_seconds(self) -> float:
        """Return the duration as seconds.

        Returns:
            float: The calculated total seconds value.
        """
        return self.duration.days * 86400 + self.duration.seconds

    @property
    def computed_total_seconds(self) -> float:
        """Return the computed_total_seconds as seconds.

        Returns:
            float: The computed_total_seconds value.
        """
        return self.computed_duration.days * 86400 + self.computed_duration.seconds


class ProjectMixin(object):
    """Allows connecting a :class:`.Project` to the mixed in object.

    This also forces a ``all, delete-orphan`` cascade, so when a :class:``.Project``
    instance is deleted then all the class instances that are inherited from
    ``ProjectMixin`` will also be deleted. Meaning that, a class which also derives from
    ``ProjectMixin`` will not be able to exists without a project (``delete-orphan``
    case).

    Args:
        project (Project): A :class:`.Project` instance holding the project which this
            object is related to. It cannot be None, or anything other than a
            :class:`.Project` instance.
    """

    #    # add this lines for Sphinx
    #    __tablename__ = "ProjectMixins"

    @declared_attr
    def project_id(cls) -> Mapped[Optional[int]]:
        """Create the project_id attribute as a declared attribute.

        Returns:
            Column: The Column related to the project_id attribute.
        """
        return mapped_column(
            "project_id",
            Integer,
            ForeignKey("Projects.id"),
            # cannot use nullable cause a Project object needs
            # insert itself as the project and it needs post_update
            # thus nullable should be True
        )

    @declared_attr
    def project(cls) -> Mapped[Optional["Project"]]:
        """Create the project attribute as a declared attribute.

        Returns:
            relationship: The relationship object related to the project attribute.
        """
        backref_table_name = cls.__tablename__.lower()
        doc = """The :class:`.Project` instance that this object belongs to."""

        return relationship(
            "Project",
            primaryjoin=f"{cls.__tablename__}.c.project_id==Projects.c.id",
            post_update=True,  # for project itself
            uselist=False,
            backref=backref(backref_table_name, cascade="all, delete-orphan"),
            doc=doc,
        )

    def __init__(
        self, project: Optional["Project"] = None, **kwargs: Dict[str, Any]
    ) -> None:
        self.project = project

    @validates("project")
    def _validate_project(self, key: str, project: "Project") -> "Project":
        """Validate the given project value.

        Args:
            key (str): The name of the validated column.
            project (Project): The project value to be validated.

        Raises:
            TypeError: If the project is None or not a Project instance.

        Returns:
            Project: The validated project value.
        """
        from stalker.models.project import Project

        if project is None:
            raise TypeError(
                f"{self.__class__.__name__}.project cannot be None it must be an "
                "instance of stalker.models.project.Project"
            )

        if not isinstance(project, Project):
            raise TypeError(
                f"{self.__class__.__name__}.project should be an instance of "
                "stalker.models.project.Project instance, "
                f"not {project.__class__.__name__}: '{project}'"
            )
        return project


class ReferenceMixin(object):
    """Adds reference capabilities to the mixed in class.

    References are :class:`stalker.models.file.File` instances or anything
    derived from it, which adds information to the attached objects. The aim of
    the References are generally to give more info to direct the evolution of
    the object.

    Args:
        references (File): A list of :class:`.File` instances.
    """

    # add this lines for Sphinx
    #    __tablename__ = "ReferenceMixins"

    def __init__(
        self, references: Optional[List["File"]] = None, **kwargs: Dict[str, Any]
    ) -> None:
        if references is None:
            references = []

        self.references = references

    @declared_attr
    def references(cls) -> Mapped[Optional[List["File"]]]:
        """Create the references attribute as a declared attribute.

        Returns:
            relationship: The relationship object related to the references attribute.
        """
        primary_cls_name = f"{cls.__name__}"
        secondary_cls_name = "Reference"
        primary_cls_table_name = f"{cls.__tablename__}"
        secondary_cls_table_name = "Files"
        secondary_table_name = f"{cls.__name__}_References"

        # get secondary table
        secondary_table = create_secondary_table(
            primary_cls_name=primary_cls_name,
            secondary_cls_name=secondary_cls_name,
            primary_cls_table_name=primary_cls_table_name,
            secondary_cls_table_name=secondary_cls_table_name,
            secondary_table_name=secondary_table_name,
        )
        # return the relationship
        return relationship(
            secondary=secondary_table,
            primaryjoin=f"{primary_cls_table_name}.c.id=={secondary_table_name}.c.{primary_cls_name.lower()}_id",
            secondaryjoin=f"{secondary_table_name}.c.{secondary_cls_name.lower()}_id=={secondary_cls_table_name}.c.id",
            doc="""A list of :class:`.File` instances given as a reference for
            this entity.
            """,
        )

    @validates("references")
    def _validate_references(self, key: str, reference: "File") -> "File":
        """Validate the given reference.

        Args:
            key (str): The name of the validated column.
            reference (File): The reference value to be validated.

        Raises:
            TypeError: If the reference is not a File instance.

        Returns:
            File: The validated reference value.
        """
        from stalker.models.file import File

        # all items should be instance of stalker.models.entity.Entity
        if not isinstance(reference, File):
            raise TypeError(
                f"{self.__class__.__name__}.references should only contain "
                "instances of stalker.models.file.File, "
                f"not {reference.__class__.__name__}: '{reference}'"
            )
        return reference


class ACLMixin(object):
    """A Mixin for adding ACLs to mixed in class.

    Access control lists or ACLs are used to determine if the given resource has the
    permission to access the given data. It is based on Pyramids Authorization system
    but organized to fit in Stalker style.

    The ACLMixin adds an attribute called ``permissions`` and a property called
    ``__acl__`` to be able to pass the permission data to Pyramid framework.
    """

    @declared_attr
    def permissions(cls) -> Mapped[List["Permission"]]:
        """Create the permissions attribute as a declared attribute.

        Returns:
            relationship: The relationship object related to the permissions attribute.
        """
        # get the secondary table
        secondary_table = create_secondary_table(
            cls.__name__, "Permission", cls.__tablename__, "Permissions"
        )
        return relationship("Permission", secondary=secondary_table)

    @validates("permissions")
    def _validate_permissions(self, key: str, permission: "Permission") -> "Permission":
        """Validate the given permission value.

        Args:
            key (str): The name of the validated column.
            permission (Permission): The permission value to be validated.

        Raises:
            TypeError: If the given permission value is not a Permission instance.

        Returns:
            Permission: The validated permission value.
        """
        from stalker.models.auth import Permission

        if not isinstance(permission, Permission):
            raise TypeError(
                f"{self.__class__.__name__}.permissions should be all instances of "
                "stalker.models.auth.Permission, "
                f"not {permission.__class__.__name__}: '{permission}'"
            )

        return permission

    @property
    def __acl__(self) -> List[Tuple[str, str, str]]:
        """Return Pyramid friendly ACL list.

        The ACL list is composed by the:

          * Permission.access (Ex: 'Allow' or 'Deny')
          * The Mixed in class name and the object name (Ex: 'User:eoyilmaz')
          * The Action and the target class name (Ex: 'Create_Asset')

        Thus, a list of tuple is returned as follows::

          __acl__ = [
              ('Allow', 'User:eoyilmaz', 'Create_Asset'),
          ]

        For the last example user eoyilmaz can grant access to views requiring
        'Add_Project' permission.

        Returns:
            List[Tuple[str, str, str]]: A list of tuples containing the ACL .
        """
        return [
            (
                perm.access,
                f"{self.__class__.__name__}:{self.name}",
                f"{perm.action}_{perm.class_name}",
            )
            for perm in self.permissions
        ]


class CodeMixin(object):
    """Adds code info to the mixed in class.

    .. versionadded:: 0.2.0

      The code attribute of the SimpleEntity is now introduced as a separate
      mixin. To let it be used by the classes it is really needed.

    The CodeMixin just adds a new field called ``code``. It is a very simple attribute
    and is used for simplifying long names (like Project.name etc.).

    Contrary to previous implementations the code attribute is not formatted in any way,
    so care needs to be taken if the code attribute is going to be used in filesystem as
    file and directory names.

    Args:
        code (str): The code attribute is a string, cannot be empty or cannot be None.
    """

    def __init__(
        self,
        code: Optional[str] = None,
        **kwargs: Dict[str, Any],
    ) -> None:
        logger.debug(f"code: {code}")
        self.code = code

    @declared_attr
    def code(cls) -> Mapped[str]:
        """Create the code attribute as a declared attribute.

        Returns:
            Column: The Column related to the code attribute.
        """
        return mapped_column(
            "code",
            String(256),
            nullable=False,
            doc="""The code name of this object.

                It accepts strings. Cannot be None.""",
        )

    @validates("code")
    def _validate_code(self, key: str, code: str) -> str:
        """Validate the given code attribute.

        Args:
            key (str): The name of the validated column.
            code (str): The code value to be validated.

        Raises:
            TypeError: If the given code is not a str.
            ValueError: If the given code value is an empty str.

        Returns:
            str: The validated code value.
        """
        logger.debug(f"validating code value of: {code}")
        if code is None:
            raise TypeError(f"{self.__class__.__name__}.code cannot be None")

        if not isinstance(code, str):
            raise TypeError(
                f"{self.__class__.__name__}.code should be a string, "
                f"not {code.__class__.__name__}: '{code}'"
            )

        if code == "":
            raise ValueError(
                f"{self.__class__.__name__}.code cannot be an empty string"
            )

        return code


class WorkingHoursMixin(object):
    """Set working hours for the mixed in class.

    Generally is meaningful for users, departments and studio.

    Args:
        working_hours (WorkingHours): A :class:`.WorkingHours` instance showing the
            working hours settings.
    """

    def __init__(
        self, working_hours: Optional["WorkingHours"] = None, **kwargs: Dict[str, Any]
    ) -> None:
        self.working_hours = working_hours

    @declared_attr
    def working_hours_id(cls) -> Mapped[Optional[int]]:
        """Create the working_hours_id attribute as a declared attribute.

        Returns:
            Column: The Column related to the working_hours_id attribute.
        """
        return mapped_column("working_hours_id", Integer, ForeignKey("WorkingHours.id"))

    @declared_attr
    def working_hours(cls) -> Mapped[Optional["WorkingHours"]]:
        """Create the working_hours attribute as a declared attribute.

        Returns:
            relationship: The relationship object related to the working_hours
                attribute.
        """
        return relationship(
            "WorkingHours",
            primaryjoin=f"{cls.__name__}.working_hours_id==WorkingHours.working_hours_id",
        )

    @validates("working_hours")
    def _validate_working_hours(
        self, key, wh: Union[None, "WorkingHours"]
    ) -> "WorkingHours":
        """Validate the given working hours value.

        Args:
            key (str): The name of the validated column.
            wh (WorkingHours): The working hours value to be validated.

        Raises:
            TypeError: If the working hours is not None and not a WorkingHours instance.

        Returns:
            WorkingHours: The validated WorkingHours value.
        """
        from stalker.models.studio import WorkingHours

        if wh is None:
            wh = WorkingHours()  # without any argument this will use the
            # default.working_hours settings
        elif not isinstance(wh, WorkingHours):
            raise TypeError(
                f"{self.__class__.__name__}.working_hours should be a "
                "stalker.models.studio.WorkingHours instance, "
                f"not {wh.__class__.__name__}: '{wh}'"
            )

        return wh


class ScheduleMixin(object):
    """Add schedule info to the mixed in class.

    Add attributes like schedule_timing, schedule_unit and schedule_model
    attributes to the mixed in class.

    Use the ``__default_schedule_attr_name__`` attribute to customize the
    column names.
    """

    # some default values that can be overridden in Mixed in classes
    __default_schedule_attr_name__ = "schedule"
    __default_schedule_timing__ = defaults.timing_resolution.seconds / 60
    __default_schedule_unit__ = TimeUnit.Hour
    __default_schedule_model__ = ScheduleModel.Effort

    def __init__(
        self,
        schedule_timing: Optional[float] = None,
        schedule_unit: TimeUnit = TimeUnit.Hour,
        schedule_model: Optional[ScheduleModel] = ScheduleModel.Effort,
        schedule_constraint: ScheduleConstraint = ScheduleConstraint.NONE,
        **kwargs: Dict[str, Any],
    ) -> None:
        self.schedule_constraint = schedule_constraint
        self.schedule_model = schedule_model
        self.schedule_timing = schedule_timing
        self.schedule_unit = schedule_unit

    @declared_attr
    def schedule_timing(cls) -> Mapped[Optional[float]]:
        """Create the schedule_timing attribute as a declared attribute.

        Returns:
            Column: The Column related to the schedule_timing attribute.
        """
        return mapped_column(
            f"{cls.__default_schedule_attr_name__}_timing",
            Float,
            nullable=True,
            default=0,
            doc="""It is the value of the {attr} timing. It is a float value.

            The timing value can either be as Work Time or Calendar Time
            defined by the {attr}_model attribute. So when the {attr}_model
            is `duration` then the value of this attribute is in Calendar Time,
            and if the {attr}_model is either `length` or `effort` then the
            value is considered as Work Time.
            """.format(
                attr=cls.__default_schedule_attr_name__
            ),
        )

    @declared_attr
    def schedule_unit(cls) -> Mapped[Optional[TimeUnit]]:
        """Create the schedule_unit attribute as a declared attribute.

        Returns:
            Column: The Column related to the schedule_unit attribute.
        """
        return mapped_column(
            f"{cls.__default_schedule_attr_name__}_unit",
            TimeUnitDecorator(),
            nullable=True,
            default=TimeUnit.Hour,
            doc=f"It is the unit of the {cls.__default_schedule_attr_name__} "
            "timing. It is a TimeUnit enum value.",
        )

    @declared_attr
    def schedule_model(cls) -> Mapped[ScheduleModel]:
        """Create the schedule_model attribute as a declared attribute.

        Returns:
            Column: The Column related to the schedule_model attribute.
        """
        return mapped_column(
            f"{cls.__default_schedule_attr_name__}_model",
            ScheduleModelDecorator(),
            default=ScheduleModel.Effort,
            nullable=False,
            doc="""Defines the schedule model which is used by **TaskJuggler**
            while scheduling this Projects. It is handled as a ScheduleModel
            enum value which has three possible values; **effort**,
            **duration**, **length**. :attr:`.ScheduleModel.Effort` is the
            default value. Each value causes this task to be scheduled in
            different ways:

            ======== ==========================================================
            effort   If the :attr:`.schedule_model` attribute is set to
                     **"effort"** then the start and end date values are
                     calculated so that a resource should spent this much of
                     work time to complete a Task. For example, a task with
                     :attr:`.schedule_timing` of 4 days, needs 4 working days.
                     So it can take 4 working days to complete the Task, but it
                     doesn't mean that the task duration will be 4 days. If the
                     resource works overtime then the task will be finished
                     before 4 days or if the resource will not be available
                     (due to a vacation or task coinciding to a weekend day)
                     then the task duration can be much more bigger than
                     required effort.

            duration The duration of the task will exactly be equal to
                     :attr:`.schedule_timing` regardless of the resource
                     availability. So the difference between :attr:`.start`
                     and :attr:`.end` attribute values are equal to
                     :attr:`.schedule_timing`. Essentially making the task
                     duration in calendar days instead of working days.

            length   In this model the duration of the task will exactly be
                     equal to the given length value in working days regardless
                     of the resource availability. So a task with the
                     :attr:`.schedule_timing` is set to 4 days will be
                     completed in 4 working days. But again it will not be
                     always 4 calendar days due to the weekends or non working
                     days.
            ======== ==========================================================
            """,
        )

    @declared_attr
    def schedule_constraint(cls) -> Mapped[ScheduleConstraint]:
        """Create the schedule_constraint attribute as a declared attribute.

        Returns:
            Column: The Column related to the schedule_constraint attribute.
        """
        return mapped_column(
            f"{cls.__default_schedule_attr_name__}_constraint",
            ScheduleConstraintDecorator(),
            default=0,
            nullable=False,
            doc="""A ScheduleConstraint value showing the constraint schema
            for this task.

            Possible values are:

             ===== ===============
               0   Constrain None
               1   Constrain Start
               2   Constrain End
               3   Constrain Both
             ===== ===============

            This value is going to be used to constrain the start and end date
            values of this task. So if you want to pin the start of a task to a
            certain date. Set its :attr:`.schedule_constraint` value to
            :attr:`.ScheduleConstraint.Start`. When the task is scheduled by
            **TaskJuggler** the start date will be pinned to the :attr:`start`
            attribute of this task.

            And if both of the date values (start and end) wanted to be pinned
            to certain dates (making the task effectively a ``duration`` task)
            set the desired :attr:`start` and :attr:`end` and then set the
            :attr:`schedule_constraint` to :att:`.ScheduleConstraint.Both`.
            """,
        )

    @validates("schedule_constraint")
    def _validate_schedule_constraint(
        self,
        key: str,
        schedule_constraint: Union[None, int, str],
    ) -> ScheduleConstraint:
        """Validate the given schedule_constraint value.

        Args:
            key (str): The name of the validated column.
            schedule_constraint (Union[None, int, str]): The value to be
                validated.

        Returns:
            ScheduleConstraint: The validated schedule_constraint value.
        """
        if schedule_constraint is None:
            schedule_constraint = ScheduleConstraint.NONE

        schedule_constraint = ScheduleConstraint.to_constraint(schedule_constraint)

        return schedule_constraint

    @validates("schedule_model")
    def _validate_schedule_model(
        self, key: str, schedule_model: Union[None, str, ScheduleModel]
    ) -> ScheduleModel:
        """Validate the given schedule_model value.

        Args:
            key (str): The name of the validated column.
            schedule_model (Union[None, str]): The schedule_model value to be
                validated.

        Returns:
            ScheduleModel: The validated schedule_model value.
        """
        if schedule_model is None:
            schedule_model = self.__default_schedule_model__
        else:
            schedule_model = ScheduleModel.to_model(schedule_model)

        return schedule_model

    @validates("schedule_unit")
    def _validate_schedule_unit(
        self, key: str, schedule_unit: Union[None, str, TimeUnit]
    ) -> TimeUnit:
        """Validate the given schedule_unit.

        Args:
            key (str): The name of the validated column.
            schedule_unit (Union[None, str, TimeUnit]): The schedule_unit value
                to be validated.

        Returns:
            TimeUnit: The validated schedule_unit value.
        """
        if schedule_unit is None:
            schedule_unit = self.__default_schedule_unit__

        schedule_unit = TimeUnit.to_unit(schedule_unit)

        return schedule_unit

    @validates("schedule_timing")
    def _validate_schedule_timing(
        self,
        key: str,
        schedule_timing: Union[None, int, float],
    ) -> float:
        """Validate the given schedule_timing.

        Args:
            key (str): The name of the validated column.
            schedule_timing (Union[None, int, float]): The schedule_timing
                value to be validated.

        Raises:
            TypeError: If the given schedule_timing is not an int or float.

        Returns:
            float: The validated schedule_timing value.
        """
        if schedule_timing is None:
            schedule_timing = self.__default_schedule_timing__
            self.schedule_unit = self.__default_schedule_unit__

        if not isinstance(schedule_timing, (int, float)):
            raise TypeError(
                "{cls}.{attr}_timing should be an integer or float "
                "number showing the value of the {attr} timing of this "
                "{cls}, not {timing_class}: '{timing}'".format(
                    cls=self.__class__.__name__,
                    attr=self.__default_schedule_attr_name__,
                    timing_class=schedule_timing.__class__.__name__,
                    timing=schedule_timing,
                )
            )

        return schedule_timing

    @classmethod
    def least_meaningful_time_unit(
        cls, seconds: int, as_work_time: bool = True
    ) -> Tuple[int, TimeUnit]:
        """Return the least meaningful time unit that corresponds to the given seconds.

        So if:

          as_work_time == True
              seconds % (1 year work time as seconds) == 0 --> 'y' else:
              seconds % (1 month work time as seconds) == 0 --> 'm' else:
              seconds % (1 week work time as seconds) == 0 --> 'w' else:
              seconds % (1 day work time as seconds) == 0 --> 'd' else:
              seconds % (1 hour work time as seconds) == 0 --> 'h' else:
              seconds % (1 minute work time as seconds) == 0 --> 'min' else:
              raise RuntimeError
          as_work_time == False
              seconds % (1 years as seconds) == 0 --> 'y' else:
              seconds % (1 month as seconds) == 0 --> 'm' else:
              seconds % (1 week as seconds) == 0 --> 'w' else:
              seconds % (1 day as seconds) == 0 --> 'd' else:
              seconds % (1 hour as seconds) == 0 --> 'h' else:
              seconds % (1 minutes as seconds) == 0 --> 'min' else:
              raise RuntimeError

        Args:
            seconds (int): An integer showing the total seconds to be
                converted.
            as_work_time (bool): Should the input be considered as work time or
                calendar time.

        Returns:
            int, TimeUnit: Returns one integer and a TimeUnit enum value,
                showing the timing value and the unit.
        """
        minutes = 60
        hour = 3600
        day = 86400
        week = 604800
        month = 2419200
        year = 31536000

        day_wt = defaults.daily_working_hours * 3600
        week_wt = defaults.weekly_working_days * day_wt
        month_wt = 4 * week_wt
        year_wt = int(defaults.yearly_working_days) * day_wt

        if as_work_time:
            logger.debug("calculating in work time")
            if seconds % year_wt == 0:  # noqa: S001
                return seconds // year_wt, TimeUnit.Year
            elif seconds % month_wt == 0:  # noqa: S001
                return seconds // month_wt, TimeUnit.Month
            elif seconds % week_wt == 0:  # noqa: S001
                return seconds // week_wt, TimeUnit.Week
            elif seconds % day_wt == 0:  # noqa: S001
                return seconds // day_wt, TimeUnit.Day
        else:
            logger.debug("calculating in calendar time")  # noqa: S001
            if seconds % year == 0:  # noqa: S001
                return seconds // year, TimeUnit.Year
            elif seconds % month == 0:  # noqa: S001
                return seconds // month, TimeUnit.Month
            elif seconds % week == 0:  # noqa: S001
                return seconds // week, TimeUnit.Week
            elif seconds % day == 0:  # noqa: S001
                return seconds // day, TimeUnit.Day

        # in either case
        if seconds % hour == 0:  # noqa: S001
            return seconds // hour, TimeUnit.Hour

        # at this point we understand that it has a residual of less then one
        # minute so return in minutes
        return seconds // minutes, TimeUnit.Minute

    @classmethod
    def to_seconds(
        cls,
        timing: float,
        unit: Union[None, str, TimeUnit],
        model: Union[str, ScheduleModel],
    ) -> Union[None, float]:
        """Convert the schedule values to seconds.

        Depending on to the schedule_model the value will differ. So if the
        schedule_model is 'effort' or 'length' then the schedule_time and schedule_unit
        values are interpreted as work time, if the schedule_model is 'duration' then
        the schedule_time and schedule_unit values are considered as calendar time.

        Args:
            timing (float): The timing value.
            unit (Union[None, str, TimeUnit]): The unit value, a TimeUnit enum
                value or one of ['min', 'h', 'd', 'w', 'm', 'y', 'Minute',
                'Hour', 'Day', 'Week', 'Month', 'Year'].
            model (str): The schedule model, preferably a ScheduleModel enum
                value or one of 'effort', 'length' or 'duration'.

        Returns:
            Union[None, float]: The converted seconds value.
        """
        if not unit:
            return None

        unit = TimeUnit.to_unit(unit)

        lut = {
            TimeUnit.Minute: 60,
            TimeUnit.Hour: 3600,
            TimeUnit.Day: 86400,
            TimeUnit.Week: 604800,
            TimeUnit.Month: 2419200,
            TimeUnit.Year: 31536000,
        }

        if model in [ScheduleModel.Effort, ScheduleModel.Length]:
            day_wt = defaults.daily_working_hours * 3600
            week_wt = defaults.weekly_working_days * day_wt
            month_wt = 4 * week_wt
            year_wt = int(defaults.yearly_working_days) * day_wt

            lut = {
                TimeUnit.Minute: 60,
                TimeUnit.Hour: 3600,
                TimeUnit.Day: day_wt,
                TimeUnit.Week: week_wt,
                TimeUnit.Month: month_wt,
                TimeUnit.Year: year_wt,
            }

        return timing * lut[unit]

    @classmethod
    def to_unit(
        cls,
        seconds: int,
        unit: Union[None, str, TimeUnit],
        model: Union[str, ScheduleModel],
    ) -> float:
        """Convert the ``seconds`` value to the given ``unit``.

        Depending on to the ``schedule_model`` the value will differ. So if the
        ``schedule_model`` is 'effort' or 'length' then the ``seconds`` and
        ``schedule_unit`` values are interpreted as work time, if the
        ``schedule_model`` is :attr:`ScheduleModel.Duration` then the
        ``seconds`` and ``schedule_unit`` values are considered as calendar
        time.

        Args:
            seconds (int): The seconds to convert.
            unit (Union[None, str, TimeUnit]): The unit value, a TimeUnit enum
                value one of ['min', 'h', 'd', 'w', 'm', 'y', 'Minute', 'Hour',
                'Day', 'Week', 'Month', 'Year'] or a TimeUnit enum value.
            model (Union[str, ScheduleModel]): The schedule model, either a
                ScheduleModel enum value or one of 'effort', 'length' or
                'duration'.

        Returns:
            float: The seconds converted to the given unit considering the
                given model.
        """
        if unit is None:
            return None

        unit = TimeUnit.to_unit(unit)
        model = ScheduleModel.to_model(model)

        lut = {
            TimeUnit.Minute: 60,
            TimeUnit.Hour: 3600,
            TimeUnit.Day: 86400,
            TimeUnit.Week: 604800,
            TimeUnit.Month: 2419200,
            TimeUnit.Year: 31536000,
        }

        if model in [ScheduleModel.Effort, ScheduleModel.Length]:
            day_wt = defaults.daily_working_hours * 3600
            week_wt = defaults.weekly_working_days * day_wt
            month_wt = 4 * week_wt
            year_wt = int(defaults.yearly_working_days) * day_wt

            lut = {
                TimeUnit.Minute: 60,
                TimeUnit.Hour: 3600,
                TimeUnit.Day: day_wt,
                TimeUnit.Week: week_wt,
                TimeUnit.Month: month_wt,
                TimeUnit.Year: year_wt,
            }

        return seconds / lut[unit]

    @property
    def schedule_seconds(self) -> float:
        """Return the schedule values as seconds.

        Depending on to the schedule_model the value will differ. So if the
        schedule_model is 'effort' or 'length' then the schedule_time and schedule_unit
        values are interpreted as work time, if the schedule_model is 'duration' then
        the schedule_time and schedule_unit values are considered as calendar time.

        Returns:
            float: The schedule_seconds as seconds.
        """
        return self.to_seconds(
            self.schedule_timing, self.schedule_unit, self.schedule_model
        )


class DAGMixin(object):
    """DAG mixin adds attributes required for parent/child relationship.

    Create a parent/child or a directed acyclic graph (DAG) relation on the mixed in
    class by introducing two new attributes called parent and children.

    Please set the ``__id_column__`` attribute to the id column of the mixed in
    class to be able to use this mixin::

    .. code-block: python

        class MixedInClass(SomeBaseClass, DAGMixin):

            id : Mapped[int] = mapped_column('id', primary_key=True)
            __id_column__ = id

    Use the :attr:``.__dag_cascade__`` to control the cascade behavior.
    """

    __dag_cascade__ = "all, delete"

    @declared_attr
    def parent_id(cls) -> Mapped[Optional[int]]:
        """Create the parent_id attribute as a declared attribute.

        Returns:
            Column: The Column related to the parent_id attribute.
        """
        return mapped_column(
            "parent_id", Integer, ForeignKey(f"{cls.__tablename__}.id")
        )

    @declared_attr
    def parent(cls) -> Mapped[Self]:
        """Create the parent attribute as a declared attribute.

        Returns:
            relationship: The relationship object related to the parent attribute.
        """
        return relationship(
            cls.__name__,
            remote_side=[getattr(cls, cls.__id_column__)],
            primaryjoin="{ct}.c.parent_id=={ct}.c.id".format(ct=cls.__tablename__),
            back_populates="children",
            post_update=True,
            doc="""A :class:`{c}` instance which is the parent of this {c}.
            In Stalker it is possible to create a hierarchy of {c}.
            """.format(
                c=cls.__name__
            ),
        )

    @declared_attr
    def children(cls) -> Mapped[List[Self]]:
        """Create the children attribute as a declared attribute.

        Returns:
            relationship: The relationship object related to the parent attribute.
        """
        return relationship(
            cls.__name__,
            primaryjoin="{ct}.c.id=={ct}.c.parent_id".format(ct=cls.__tablename__),
            back_populates="parent",
            post_update=True,
            cascade=cls.__dag_cascade__,
            doc="""Other :class:`Budget` instances which are the children of this
            one. This attribute along with the :attr:`.parent` attribute is used in
            creating a DAG hierarchy of tasks.
            """,
        )

    def __init__(self, parent: Optional[Self] = None, **kwargs: Dict[str, Any]) -> None:
        self.parent = parent

    @validates("parent")
    def _validate_parent(
        self, key: str, parent: Union[None, Self]
    ) -> Union[None, Self]:
        """Validate the given parent value.

        Args:
            key (str): The name of the validated column.
            parent (object): The parent object to be validated.

        Raises:
            TypeError: If the parent is not None and not deriving from the same class
                with this instance.

        Returns:
            Union[None, Self]: The validated parent value.
        """
        if parent is None:
            return parent

        if not isinstance(parent, self.__class__):
            raise TypeError(
                "{cls}.parent should be an instance of {cls} class or "
                "derivative, not {parent_cls}: '{parent}'".format(
                    cls=self.__class__.__name__,
                    parent_cls=parent.__class__.__name__,
                    parent=parent,
                )
            )
        check_circular_dependency(self, parent, "children")

        return parent

    @validates("children")
    def _validate_children(self, key: str, child: Self) -> Self:
        """Validate the given child.

        Args:
            key (str): The name of the validated column.
            child (Self): The child value to be validated.

        Raises:
            TypeError: If any of the child objects are not deriving from the same class
                as this one.

        Returns:
            Self: The validated child instance.
        """
        if not isinstance(child, self.__class__):
            raise TypeError(
                "{cls}.children should only contain instances of {cls} "
                "(or derivative), not {child_cls}: '{child}'".format(
                    cls=self.__class__.__name__,
                    child_cls=child.__class__.__name__,
                    child=child,
                )
            )

        return child

    @property
    def is_root(self) -> bool:
        """Return True if the Task has no parent.

        Returns:
            bool: True if the Task has no parent.
        """
        return not bool(self.parent)

    @property
    def is_container(self) -> bool:
        """Return True if the Task has children Tasks.

        Returns:
            bool: True if the Task has children Tasks.
        """
        with DBSession.no_autoflush:
            return bool(len(self.children))

    @property
    def is_leaf(self) -> bool:
        """Return True if the Task has no children Tasks.

        Returns:
            bool: True if the Task has no children Tasks.
        """
        return not self.is_container

    @property
    def parents(self) -> List[Self]:
        """Return all of the parents of this mixed in class starting from the root.

        Returns:
            List[Self]: List of tasks showing the parent of this Task.
        """
        parents = []
        entity = self.parent
        # TODO: make this a generator
        while entity:
            parents.append(entity)
            entity = entity.parent
        parents.reverse()
        return parents

    def walk_hierarchy(
        self,
        method: Union[int, str, TraversalDirection] = TraversalDirection.DepthFirst,
    ) -> Generator[None, Self, None]:
        """Walk the hierarchy of this task.

        Args:
            method (Union[int, str, TraversalDirection]): The walk method
                defined by the :class:`.TraversalDirection` enum value. The
                default is :attr:`.TraversalDirection.DepthFirst`.

        Yields:
            Task: The child Task.
        """
        for c in walk_hierarchy(self, "children", method=method):
            yield c


class AmountMixin(object):
    """Adds ``amount`` attribute to the mixed in class.

    Args:
        amount (Union[int, float]): The amount value.
    """

    def __init__(self, amount: Union[int, float] = 0, **kwargs: Dict[str, Any]) -> None:
        self.amount = amount

    @declared_attr
    def amount(cls) -> Mapped[Optional[float]]:
        """Create the amount attribute as a declared attribute.

        Returns:
            Column: The Column related to the amount attribute.
        """
        return mapped_column(Float, default=0.0)

    @validates("amount")
    def _validate_amount(self, key: str, amount: Union[int, float]) -> float:
        """Validate the given amount value.

        Args:
            key (str): The name of the validated column.
            amount (Union[int, float]): The amount value to be validated.

        Raises:
            TypeError: If the given amount value is not a int or float.

        Returns:
            float: The validated amount value.
        """
        if amount is None:
            amount = 0.0

        if not isinstance(amount, (int, float)):
            raise TypeError(
                f"{self.__class__.__name__}.amount should be a number, "
                f"not {amount.__class__.__name__}: '{amount}'"
            )

        return float(amount)


class UnitMixin(object):
    """Adds ``unit`` attribute to the mixed in class.

    Args:
        unit (str): The unit of this mixed in class.
    """

    def __init__(self, unit: str = "", **kwargs: Dict[str, Any]) -> None:
        self.unit = unit

    @declared_attr
    def unit(cls) -> Mapped[Optional[str]]:
        """Create the unit attribute as a declared attribute.

        Returns:
            Column: The Column related to the unit attribute.
        """
        return mapped_column(String(64))

    @validates("unit")
    def _validate_unit(self, key: str, unit: Union[None, str]) -> str:
        """Validate the given unit value.

        Args:
            key (str): The name of the validated column.
            unit (str): The unit value to be validated.

        Raises:
            TypeError: If the given unit is not a str.

        Returns:
            str: The validated unit value.
        """
        if unit is None:
            unit = ""

        if not isinstance(unit, str):
            raise TypeError(
                f"{self.__class__.__name__}.unit should be a string, "
                f"not {unit.__class__.__name__}: '{unit}'"
            )

        return unit
